#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import sys

try:
    from bs4 import BeautifulSoup
except ImportError as e:
    print(
        "\n  ".join([
            "Please install bs4",
            "APT: sudo apt install python3-bs4",
            "PIP: sudo pip3 install bs4",
        ]),
        file = sys.stderr
    )
    raise e

import difflib, operator, re
from pprint             import pformat, pprint
from .binary_predicate  import BinaryPredicate
from .connector         import Connector
from .download          import download
from .log               import Log
from .query             import ACTION_READ, Query
from .scholar           import ScholarArticle, ScholarConf, ScholarQuerier, ScholarSettings, SearchScholarQuery, SoupKitchen

def parse_article(s_html :str) -> dict:
    """
    Parse a "gs_res_ccl_mid" div (wrapping each article) returned by Google Scholar.
    Args:
        s_html: The HTML string containing the div.
    Returns:
        The dict describing the article, structured as follows:
        {
            "authors"       : list, # The list only contains the first authors. The last author name maybe incomplete.
            "cluster_id"    : int,
            "conference"    : str,  # May be incomplete
            "editor"        : str,
            "excerpt"       : str,  # May be incomplete.
            "num_citations" : int,
            "num_versions"  : int,
            "title"         : str,
            "url_citations" : str,
            "url_versions"  : str,
            "url_pdf"       : str,
            "url_title"     : str,
            "year"          : int,
        }
        An incomplete string may start by '…' and ends with '…'.
        URLs are absolute.
    """
    def clean_string(s :str) -> str:
        return re.sub("( |\xa0|\n)+", " ", s).strip()

    def extract_first_int(s :str) -> int:
        return int(re.search("\d+", s).group())

    entry = dict()
    soup = BeautifulSoup(s_html, features = "lxml")

    # Extract url_title
    div = soup.find(name="div", attrs={"class" : "gs_or_ggsm"})
    if div:
        a = div.find(name="a")
        if a:
            entry["url_pdf"] = a["href"]

    # Extract url_title, title
    h3 = soup.find(name="h3", attrs={"class" : "gs_rt"})
    if h3:
        span = h3.find(name="span", attrs={"class" : "gt_ct1"})
        if span: # [CITATION]
            entry["doc_type"] = span.text[1:-1]
        a = h3.find(name="a")
        if a:
            entry["url_title"] = a["href"]
            entry["title"] = clean_string(a.text)

    # Extract authors, year, editor
    div = soup.find(name="div", attrs={"class" : "gs_a"})
    if div:
        s = div.text.replace("\xa0", " ")
        (authors, conference_year, editor) = s.split(" - ")
        entry["authors"] = [clean_string(a) for a in authors.split(",")]
        conference_year = [clean_string(elt) for elt in conference_year.split(",")]
        entry["conference"] = conference_year[0] if len(conference_year) == 2 else None
        entry["year"] = int(conference_year[-1])
        entry["editor"] = clean_string(editor)

    # Extract exceprt
    div = soup.find(name="div", attrs={"class" : "gs_rs"})
    if div:
        entry["excerpt"] = clean_string(div.text.strip())

    # Extract num_citations and num_versions
    divs = soup.findAll(name="div", attrs={"class" : "gs_fl"})
    if divs and len(divs) > 1:
        links = divs[1].findAll("a")
        if links:
            # It is important to not use link index here as it may change depending on the
            # document (e.g whether it is cited or not).
            for link in links:
                s =  link.text.lower()
                key = "citations" if "cited by" in s else \
                      "versions"  if "versions" in s else \
                      None
                if key:
                    entry["num_%s" % key] = extract_first_int(s)
                    entry["url_%s" % key] = ScholarConf.SCHOLAR_SITE + link["href"]
                    entry["cluster_id"]    = extract_first_int(link["href"])
    return entry

class MinifoldScholarQuerier(ScholarQuerier):
    """
    ScholarQuerier is overloaded to fetch more attributes.
    """
    def __init__(self):
        super().__init__()
        self.articles = list()

    def parse(self, s_html :str):
        soup = SoupKitchen.make_soup(s_html)
        soup = soup.find(name="div", attrs={"id" : "gs_res_ccl_mid"})
        for div in soup.findAll(name="div", attrs={"class" : "gs_r"}):
            s = div.prettify()
            entry = parse_article(s)
            self.articles.append(entry)

    def send_query(self, gs_query):
        # Network
        url = gs_query.get_url()
        Log.info("GoogleScholar <-- %s" % url)

        # Here, we rely on minifold downloader as it can cache results.
        #s_html = self._get_http_response(url)
        response = download(url)
        s_html = response.text

        # Parsing
        assert s_html
        self.parse(s_html)
        self.articles = list()
        self.parse(s_html)

class GoogleScholarConnector(Connector):
    def __init__(self, citation_format :str = ScholarSettings.CITFORM_BIBTEX):
        settings = ScholarSettings()
        settings.set_citation_format(citation_format)
        self.querier = MinifoldScholarQuerier() # ScholarQuerier()
        self.querier.apply_settings(settings)

    def attributes(self, object :str) -> set:
        if object == "publication" or not object:
            #article = ScholarArticle()
            #return set(article.keys())
            return {
                "authors",       #: str
                "cluster_id",    #: int,
                "conference",    #: str,
                "doc_type",      #: str,
                "editor",        #: str,
                "excerpt",       #: str,
                "num_citations", #: int,
                "num_versions",  #: int,
                "title",         #: str,
                "url_citations", #: str,
                "url_versions",  #: str,
                "url_pdf",       #: str,
                "url_title",     #: str,
                "year",          #: int,
            }
        elif object == "cluster":
            raise RuntimeError("Object %s is not yet supported" % object)
        else:
            raise RuntimeError("Invalid object %s" % object)

    @staticmethod
    def filter_to_scholar(p :BinaryPredicate, gs_query :SearchScholarQuery, authors :list):
        if p.operator == operator.__and__:
            GoogleScholarConnector.filter_to_scholar(p.left, gs_query, authors)
            GoogleScholarConnector.filter_to_scholar(p.right, gs_query, authors)
            return

        if isinstance(p.left,  BinaryPredicate) \
        or isinstance(p.right, BinaryPredicate):
            raise RuntimeError("Invalid clause %s" % p)

        attr = p.left
        value = p.right
        if not isinstance(attr, str):
            raise RuntimeError("Invalid left operand %s" % p)

        if attr == "authors":
            if p.operator in {operator.__eq__, operator.__contains__}:
                authors.append(value)
                gs_query.set_author(value)
            else:
                raise RuntimeError("Invalid operator %s" % p)
        elif attr == "words":
            if p.operator == operator.__eq__ or p.operator == operator.__ge__:
                gs_query.set_words(options.allw)      # All of these words must appear
            elif p.operator == operator.__le__:
                gs_query.set_words_some(options.some) # Some of these words must appear
            elif p.operator == operator.__ne__:
                gs_query.set_words_none(options.none) # None of these words must appear
            else:
                raise RuntimeError("Invalid operator %s" % p)
        elif attr == "phrase":
            Log.warning("GoogleScholarConnector: filtering on %s not yet implemented" % attr)
            #gs_query.set_phrase(options.phrase)
            pass
        elif attr == "doc_type":
            Log.warning("GoogleScholarConnector: filtering on %s not yet implemented" % attr)
            # NOT YET IMPLEMENTED
            # gs_query.set_include_patents(False)
            # gs_query.set_include_citations(False)
            pass
        elif attr == "year":
            assert isinstance(value, int)
            (start, end) = gs_query.timeframe
            if p.operator == operator.__eq__:
                start = max(start, value) if start else value
                end   = min(end, value) if end else value
            elif p.operator == operator.__le__:
                end = min(end, value)     if end else value
            elif p.operator == operator.__lt__:
                end = min(end, value - 1) if end else value
            elif p.operator == operator.__ge__:
                start = max(start, value) if start else value
            elif p.operator == operator.__gt__:
                start = max(start, value + 1) if start else value
            else:
                raise RuntimeError("Invalid operator %s" % p)
            if start is not None and end is not None and end < start:
                raise RuntimeError("Invalid range of date")
            gs_query.set_timeframe(start, end)
        elif attr == "publication":
            if p.operator == operator.__eq__:
                gs_query.set_pub(value)
            else:
                raise RuntimeError("Invalid operator %s" % p)

    @staticmethod
    def sanitize_author(authors :list, author :str) -> str:
        """
        Find in an input list of strings the closest string with the input string.
        Args:
            authors: The list of strings.
            author: The reference string.
        Returns:
            The string of `authors` closest to `author` if any, else `author`
        """
        # Find the best match in authors. Indeed if we search "John Doe", Google Scholar
        # returns "J Doe" and so the self.reshape_entries will drop the records.
        # In practice, the cutoff must be quite low, but not too low. Otherwise,
        # co-authors may be wrongly renamed! Empirically, 0.4 is a good tradeoff).
        l = difflib.get_close_matches(author, authors, n=1, cutoff=0.4)
        if len(l) != 1:
            # We are trying to rename a co-author without knowing is real name.
            return author
        else:
            # We are renaming an author with his/her real name.
            return l[0]

    def query(self, query :Query) -> list:
        super().query(query)
        entries = None
        if query.action != ACTION_READ:
            raise RuntimeError("Action not supported" % query.action)
        if query.filters and not isinstance(query.filters, BinaryPredicate):
            raise RuntimeError("Invalid filter: %s" % query.filters)

        authors = list()
        if query.object == "cluster":
            gs_query = ClusterScholarQuery(cluster=options.cluster_id)
        elif query.object == "publication" or not query.object:
            gs_query = SearchScholarQuery()
            GoogleScholarConnector.filter_to_scholar(query.filters, gs_query, authors)
        else:
            raise RuntimeError("Invalid object %r" % query.object)

        # Craft the query
        if query.limit is not None:
            gs_query.set_num_page_results(min(query.limit, ScholarConf.MAX_PAGE_RESULTS))

        # Doc type.
        gs_query.set_include_citations(False)
        gs_query.set_include_patents(True)

        # Send the query
        self.querier.send_query(gs_query)

        # Extract results
        if isinstance(self.querier, MinifoldScholarQuerier):
            entries = self.querier.articles
        else:
            entries = [
                {k : v[0] for (k, v) in article.attrs.items()}
                for article in self.querier.articles
            ]

        if gs_query.author:
            # Sanitize author list.
            for entry in entries:
                entry["authors"] = [
                    GoogleScholarConnector.sanitize_author(authors, author)
                    for author in entry["authors"]
                ]
        entries = self.reshape_entries(query, entries)
        return self.answer(query, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector import Connector
from .html      import html
from .query     import Query
from .sort_by   import SortByConnector
from .unique    import UniqueConnector

def get_values(connector :Connector, attribute :str):
    enriched_connector = UniqueConnector(
            [attribute],
            connector
        )
    q = Query(attributes = [attribute])
    return [entry[attribute] for entry in enriched_connector.query(q)]

def show_some_values(connector :Connector, limit :int = 5, max_strlen :int = None, query :Query = None):
    if not query:
        query = Query()

    map_values = {
        attribute : get_values(connector, attribute) \
        for attribute in connector.attributes(query.object)
    }

    html(
        """
        <table>
            <tr>
                <th>Attribute</th>
                <th>#values</th>
                <th>%(limit)d first values</th>
            </tr>
            %(rows)s
        </table>
        """ % {
            "limit" : limit,
            "rows"  : "".join(["""
                <tr>
                    <td><code>%(attribute)s</code></td>
                    <td>%(num_values)d</td>
                    <td style="text-align:left"><code>%(values)s</code></td>
                </tr>
                """ % {
                    "attribute"  : attribute,
                    "num_values" : len(map_values[attribute]),
                    "values" : "<br/>".join([
                        str(x)[:max_strlen] for x in map_values[attribute][:limit]
                    ]),
                } for attribute in connector.attributes(query.object)
            ]),
        }
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from copy       import deepcopy
from .connector import Connector
from .query     import Query

def find_lambda_dependencies(func) -> set:
    entry = dict()
    while True:
        try:
            val = func(entry)
        except KeyError as e:
            needed_key = e.args[0]
            entry[needed_key] = None
            continue
        except Exception as e:
            pass
        break
    return set(entry.keys())

def find_lambdas_dependencies(map_lambdas :dict) -> dict:
    return {attr : find_lambda_dependencies(func) for attr, func in map_lambdas.items()}

def lambdas(map_lambdas :dict, entries :list, attributes :set = None) -> list:
    # Be sure that the result is deterministic without regards each
    # lambda is processed.
    # Ex: map_lambdas = {"x" : lambda e: 10 + e["x"]} is OK
    # Ex: map_lambdas = {"x" : lambda e: 10 + e["x"] + e["y"]} is OK
    # Ex: map_lambdas = {
    #       "x" : lambda e: 10 + e["x"] + e["y"]
    #       "y" : lambda e: 10 + e["y"]
    #     } is not OK because e["y"] is ambiguous when executing the lambda related to "x".
    attrs = set(map_lambdas.keys())
    if attributes:
        attrs &= set(attributes)
    for entry in entries:
        for attr, func in map_lambdas.items():
            if attr in attrs:
                try:
                    entry[attr] = func(entry)
                except KeyError:
                    entry[attr] = None
    return entries

class LambdasConnector(Connector):
    def __init__(self, map_lambdas :dict, child :Connector, map_dependencies :dict = None):
        # Ensure there is no cyclic dependency (see lambdas())
        super().__init__()
        self.m_child = child
        self.m_map_lambdas = map_lambdas

        # If no attribute dependency is provided, use find_lambda_dependencies heuristic.
        self.m_map_dependencies = map_dependencies \
            if map_dependencies is not None \
            else find_lambdas_dependencies(map_lambdas)

    @property
    def child(self) -> Connector:
        return self.m_child

    def attributes(self, object :str) -> set:
        return self.m_child.attributes(object) | set(self.m_map_lambdas.keys())

    def find_needed_attributes(self, attributes :set) -> set:
        ret = set()
        for attribute in attributes:
            needed_attributes = self.m_map_dependencies.get(attribute, set([attribute]))
            ret |= needed_attributes
        return ret

    def query(self, q :Query) -> list:
        super().query(q)
        q_child = deepcopy(q)

        if q.attributes or q.filters:
            where_attributes = self.find_needed_attributes(find_lambda_dependencies(q.filters))

            if q.attributes:
                child_attributes = self.m_child.attributes(None)
                select_attributes  = self.find_needed_attributes(set(q.attributes))
                # Remove lambda-specific attributes
                lambdas_attributes = set(self.m_map_lambdas.keys()) - {
                    dep_attr
                    for attr in q.attributes
                    for dep_attr in self.m_map_dependencies.get(attr, set())
                    if dep_attr in child_attributes
                }
                q_child.attributes = list((select_attributes | where_attributes) - lambdas_attributes)

            if where_attributes:
                # If the filter involves at least one attribute of the LambdasConnector
                # the filter is not forwarded to the child node, and the filtering is
                # achieved locally achieved locally to this LambdasConnector.
                # As a sequel, the child node may return entries that could be filtered
                # thanks to the self.reshape_entries method.
                q_child.filters = None

        entries = self.reshape_entries(
            q,
            lambdas(
                self.m_map_lambdas,
                self.child.query(q_child),
                q.attributes
            )
        )

        return self.answer(q, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import email.utils
import json, operator

from .binary_predicate      import BinaryPredicate, __in__
from .connector             import Connector
from .doc_type              import DocType
from .download              import download
from .log                   import Log
from .strings               import to_international_string, to_canonic_fullname
from .query                 import Query, ACTION_READ, SORT_ASC

# Default HAL API queried
HAL_API_URL = "https://api.archives-ouvertes.fr/search"

# Maps HAL ontology to our ontology
# Note: this is not an exhaustive list of the attributes returned by HAL.
HAL_ALIASES = {
    "en_abstract_s"         : "abstract",
    "fr_abstract_s"         : "abstract_fr",
    "authFullName_s"        : "authors",
    "bookTitle_s"           : "book",
    "citationFull_s"        : "citation_full",
    "city_s"                : "city",
    "conferenceTitle_s"     : "conference",
    "conferenceEndDateD_i"  : "conference_day",
    "conferenceEndDateM_i"  : "conference_month",
    "conferenceEndDateY_i"  : "conference_year",
    "conferenceEndDate_s"   : "conference_date",
    "conferenceOrganizer_s" : "conference_organizer",
    "country_s"             : "country_code",
    "docType_s"             : "hal_doc_type",
    "en_title_s"            : "en_title",
    "fr_title_s"            : "fr_title",
    "fileMain_s"            : "url_viewer",
    "files_s"               : "pdf",
    "journalTitle_s"        : "journal",
    "modifiedDateD_i"       : "date_modified_day",
    "modifiedDateM_i"       : "date_modified_month",
    "modifiedDateY_i"       : "date_modified_year",
    "modifiedDate_s"        : "date_modified",
    "keyword_s"             : "keywords",
    "title_s"               : "title",
    "producedDateD_i"       : "day",
    "producedDateM_i"       : "month",
    "producedDateY_i"       : "year",
    "producedDate_s"        : "date_produced",
    "releasedDateD_i"       : "date_released_day",
    "releasedDateM_i"       : "date_released_month",
    "releasedDateY_i"       : "date_released_year",
    "releasedDate_s"        : "date_released",
    "structId_i"            : "laboratory",
    "submittedDateD_i"      : "date_submitted_day",
    "submittedDateM_i"      : "date_submitted_month",
    "submittedDateY_i"      : "date_submitted_year",
    "submittedDate_s"       : "date_submitted",
    "uri_s"                 : "hal_url",
    "version_i"             : "version"
}

class HalConnector(Connector):
    def __init__(self, map_hal_id = {}, map_hal_name = {}, hal_api_url = HAL_API_URL):
        super().__init__()
        self.m_api_url      = hal_api_url
        self.m_format       = "json"
        self.m_map_hal_id   = map_hal_id
        self.m_map_hal_name = map_hal_name
        self.m_map_rev_name = {to_canonic_fullname(hal_name) : name for (name, hal_name) in map_hal_name.items()}

    def attributes(self, object :str):
        return HAL_ALIASES.keys() | {"doc_type"}

    @property
    def api_url(self) -> str:
        return self.m_api_url

    @property
    def format(self) -> str:
        return self.m_format

    @property
    def map_hal_id(self) -> dict:
        return self.m_map_hal_id

    @property
    def map_hal_name(self) -> dict:
        return self.m_map_hal_name

    @staticmethod
    def quote(s) -> str:
        return "%%22%s%%22" % s

    @staticmethod
    def string_to_hal(s) -> str:
        # Patch for Ana Busic, anyway the following would find some publications
        # TODO: declare HAL ID for her
        s = s.replace("š", "s")
        s = s.replace("ć", "c")
        # From /usr/lib/python3/dist-packages/urllib3/fields.py, see format_header_param
        s = email.utils.encode_rfc2231(s, 'utf-8').split("'")[2]
        return HalConnector.quote(s)

    @staticmethod
    def binary_predicate_to_hal(p :BinaryPredicate) -> str:
        if p.operator == operator.__and__:
            return "&fq=".join([
                HalConnector.binary_predicate_to_hal(p.left),
                HalConnector.binary_predicate_to_hal(p.right)
            ])
        if isinstance(p.left, BinaryPredicate):
            raise RuntimeError("binary_predicate_to_hal: Invalid left operand in %s: nested clauses not supported" % p)
        if isinstance(p.right, BinaryPredicate):
            raise RuntimeError("binary_predicate_to_hal: Invalid right operand in %s: nested clauses not supported" % p)
        if not isinstance(p.left, str):
            raise RuntimeError("binary_predicate_to_hal: left operand of %s must be a string" % p)

        right = p.right
        if isinstance(right, str):
            right = HalConnector.string_to_hal(right)

        if   p.operator == operator.__eq__ or p.operator == operator.__contains__:
            ret = "%s:(%s)" % (p.left, right)
        #elif p.operator == "~":
        #    ret = "%s:\"%s\"~3" % (p.left, right)
        elif p.operator == operator.__gt__:
            ret =  "%s:{%s TO *]" % (p.left, right)
        elif p.operator == operator.__lt__:
            ret = "%s:[* TO %s}" % (p.left, right)
        elif p.operator == operator.__ge__:
            ret = "%s:[%s TO *]" % (p.left, right)
        elif p.operator == operator.__le__:
            ret = "%s:[* TO %s]" % (p.left, right)
        elif p.operator == __in__:
            (start, end) = right
            ret = "%s:[%s TO %s]" % (p.left, start, end)
        else:
            raise RuntimeError("binary_predicate_to_hal: unsupported operator (%s): %s" % (p.operator, p))

        # Space in HAL operators must also be replaced
        return ret.replace(" ", "%20")

    @staticmethod
    def sanitize_dict(d :dict) -> dict:
        for k, v in d.items():
            if isinstance(v, list):
                if len(v) == 1:
                    d[k] = v[0]
                elif k == "title_s":
                    d[k] = " ".join(v)
        return d

    @staticmethod
    def to_doc_type(s :str) -> DocType:
        ret = DocType.UNKNOWN

        # TODO: use dict()
        s = s.lower()
        if   s in ["these", "ouv"]:
            ret = DocType.BOOKS_AND_THESES
        elif s in ["comm"]:
            ret = DocType.ARTICLE
        elif s == "douv":
            # "douv" means "direction d'ouvrage" (aka proceedings)
            ret = DocType.ARTICLE
        elif s == "art":
            # Unfortunately, "art" may design journal articles
            ret = DocType.JOURNAL
        elif s in ["lecture", "report", "mem", "presconf"]:
            # "mem" (mémoires) can be used for internship report. presconf = keynote
            ret = DocType.REPORT
        elif s == "poster":
            ret = DocType.POSTER
        elif s == "hdr":
            ret = DocType.HDR
        elif s in "couv":
            ret = DocType.CHAPTER
        elif s == "patent":
            ret = DocType.PATENT
        elif s in ["undefined", "other"]:
            ret = DocType.UNKNOWN
        elif s == None:
            pass
        else:
            raise ValueError("HalConnector: unknown document type %r" %s)

        return ret

    def sanitize_entry(self, entry :dict) -> dict:
        entry = HalConnector.sanitize_dict(entry)
        keys = entry.keys()

        # << WORKAROUND: Otherwise sometimes the title is not returned ?!
        if "title" not in keys:
            for title_key in ["title_en", "title_fr"]:
                try:
                    entry["title"] = entry[title_key]
                    break
                except KeyError:
                    pass
        # >> WORKAROUND

        if "docType_s" in keys:
            try:
                entry["doc_type"] = HalConnector.to_doc_type(entry["docType_s"])
            except ValueError as e:
                Log.warning("Invalid doc_type: %s" % e)
                Log.warning(entry)
                entry["doc_type"] = DocType.UNKNOWN

        if "authFullName_s" in keys:
            if isinstance(entry["authFullName_s"], str):
                entry["authFullName_s"] = [entry["authFullName_s"]]
            elif isinstance(entry["authFullName_s"], list):
                pass
            else:
                Log.warning("Invalid author list: %s", entry["authFullName_s"])

            # Convert HAL names to our names if needed.
            entry["authFullName_s"] = [
                self.m_map_rev_name.get(to_canonic_fullname(author), author) \
                for author in entry["authFullName_s"]
            ]

        return entry

    def sanitize_entries(self, entries :list) -> list:
        return [self.sanitize_entry(entry) for entry in entries]

    def query_to_hal(self, q :Query) -> str:
        if q.action != ACTION_READ:
            raise ValueError("query_to_hal: Unsupported action %s" % q.action)
        attributes = None
        if len(q.attributes) > 0:
            attributes = ",".join(q.attributes)
            if "doc_type" in q.attributes and "hal_doc_type" not in q.attributes:
                attributes += ",docType_s"
        else:
            # Note: this avoid to pull by default every fields provided by HAL.
            attributes = "*"

        object = None
        if q.object == "lincs":
            # Lab query
            object = "structId_i:(160294)"
        elif q.object == "publication":
            # Publication query
            object = "*:*"
        else:
            # This is not a standard HAL object.
            # By convention, we assume that the user is querying
            # publications related to a researcher (bibliography query).
            try:
                # Try to consider the object as a HAL ID.
                hal_id = self.map_hal_id[q.object]
                object = "*:*&fq=authIdHal_s:(%s)" % HalConnector.quote(self.map_hal_id[q.object])
            except KeyError:
                # If not found, try to translate the string using self.map_hal_name.
                # Else, use the provided name.
                fullname = q.object
                fullname = self.map_hal_name.get(fullname, fullname)
                object = "*:*&fq=authFullName_s:(%s)" % HalConnector.string_to_hal(fullname)

        if object is None:
            raise RuntimeError("Object not supported: %s" % q.object)

        url_options = [object]

        # SELECT
        if attributes:
            url_options.append("fl=%s" % attributes)

        # WHERE
        if q.filters:
            url_options.append("fq=%s" % HalConnector.binary_predicate_to_hal(q.filters))

        # OFFSET
        if q.offset:
            raise RuntimeError("%s: in query %s: OFFSET is not supported" % (self, q))

        # LIMIT
        # Hardcoded rows=2000 to guarantee that all publications are fetched.
        url_options.append("rows=%s" % (int(q.limit) if q.limit else 2000))

        # SORT
        if q.sort_by:
            url_options.append("sort=%s" % ",".join([
                "%s+%s" % (
                    attribute,
                    "asc" if sort_asc == SORT_ASC else "desc"
                ) for attribute, sort_asc in q.sort_by.items()
            ]))
        else:
            # By default, sort by descending date.
            url_options.append("sort=submittedDate_tdate+desc")

        # HalConnector expects JSON data.
        url_options.append("wt=%s" % self.format)

        q_hal = "%(server)s/?q=%(options)s" % {
            "server"  : self.api_url,
            "options" : "&".join(url_options)
        }
        return q_hal

    def query(self, q :Query) -> list:
        super().query(q)
        entries = list()
        if q.action == ACTION_READ:
            q_hal = self.query_to_hal(q)
            Log.info("--> HAL: %s" % q_hal)
            reply = download(q_hal, timeout = (2.0, 7.0))
            try:
                data = reply.text
                if self.m_format == "json":
                    data = json.loads(data)
                    try:
                        entries = self.sanitize_entries(
                            data["response"]["docs"]
                        )
                    except KeyError: # if "response" is not found, an error has occurred
                        from pprint import pformat
                        raise RuntimeError("HAL error:\n%s" % pformat(data))
                else:
                    raise RuntimeError("Format not implemented: %s" % self.m_format)
            except Exception as e:
                raise RuntimeError("Cannot get reply from %s (status %s)" % (self.m_api_url, e))

        return self.answer(q, entries)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob and Fabien Mathieu"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

# 1) API specs:
#   https://wiki.inria.fr/lincs/Dblp-api
#
# Examples:
#   https://dblp.dagstuhl.de/search/publ/api?q=fabien-mathieu&h=500&format=json
#   https://dblp.dagstuhl.de/search/publ/api?q=fabien-mathieu%20year:2014&h=500&format=json
#
# 2) By default DBLP only returns up to 30 records. See:
#   https://dblp.org/faq/How+to+use+the+dblp+search+API.html
# The default limit in DblpConnector is set to 9999.
#
# 3) It is possible to query a specific researcher using its DBLP-ID.
# The ID can be found by browsing the page related to a researcher.
#
# Example:
#   https://dblp.uni-trier.de/pers/hd/c/Chen:Chung_Shue
#   The DBLP ID can be obtained by clicking on the export bibliography icon.
#
# For the moment, only XML is supported by DBLP.
#   https://dblp.org/pid/30/1446.xml
#
try:
    import urllib3
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
except ImportError:
    raise ImportError("DblpConnector requires python3-urllib3: please run: apt-get install python3-urllib3")

try:
    import requests
except ImportError:
    raise ImportError("DblpConnector requires requests")


try:
    import xmltodict
except ImportError:
    raise ImportError("DblpConnector requires python3-xmltodict: please run: apt-get install python3-xmltodict")

import json, operator
from pprint import pformat

from .binary_predicate      import BinaryPredicate
from .connector             import Connector
from .doc_type              import DocType
from .log                   import Log
from .strings               import to_international_string, to_canonic_fullname as _to_canonic_fullname
from .query                 import Query, ACTION_READ

# Default queried DBLP API.
DBLP_API_URL = "https://dblp.dagstuhl.de"

# Maps DBLP ontology to our ontology.
DBLP_ALIASES = {
    "type"    : "dblp_doc_type",
    "venue"   : "conference",
    "url"     : "dblp_url",
}

def to_canonic_fullname(s) -> str:
    # DBLP may now returns {"@pid" : int, "text" : "Firstname Lastname"} instead of the author fullname.
    if isinstance(s, dict):
        s = s["text"]
    return _to_canonic_fullname(s)

class DblpConnector(Connector):
    def __init__(self, map_dblp_id = {}, map_dblp_name = {}, dblp_api_url = DBLP_API_URL):
        super().__init__()
        self.m_api_url = dblp_api_url
        assert not dblp_api_url.endswith("/search") and not dblp_api_url.endswith("/search/"), \
            "Invalid API URL [%s], remove '/search/' suffix" % dblp_api_url
        self.m_format  = "json" # valid values are: "xml", "json", and "jsonp".
        self.m_map_dblp_id = map_dblp_id
        self.m_map_dblp_name = map_dblp_name
        self.m_map_rev_name = {to_canonic_fullname(dblp_name) : name for (name, dblp_name) in map_dblp_name.items()}

    def attributes(self, object :str) -> set:
        return {"authors", "doc_type", "title", "type", "venue", "url", "year"} # Non exhaustive

    @property
    def api_url(self) -> str:
        return self.m_api_url

    @property
    def format(self) -> str:
        return self.m_format

    @property
    def map_dblp_name(self) -> str:
        return self.m_map_dblp_name

    @property
    def map_dblp_id(self) -> str:
        return self.m_map_dblp_id

    @staticmethod
    def to_dblp_name(s :str, map_dblp_name = {}) -> str:
        ret = map_dblp_name.get(s, s)
        s = to_international_string(s)
        words = s.lower().split()
        try:
            words.remove("")
        except ValueError:
            pass
        # a-b means that b must follows a
        # a_b means that both a and b must be found
        # $ corresponds to exact match
        # See: http://dblp.uni-trier.de/db/about/author.html
        ret = "-".join(words) + "$"

        return ret

    def get_dblp_name(self, s :str) -> str:
        return DblpConnector.to_dblp_name(s, self.map_dblp_name)

    def get_dblp_id(self, s :str) -> str:
        pid = self.m_map_dblp_id.get(s)
        return pid if pid else self.get_dblp_name(s)

    def binary_predicate_to_dblp(self, p :BinaryPredicate, result :dict):
        # Recursive call only supported for && clauses
        if p.operator == operator.__and__:
            self.binary_predicate_to_dblp(p.left, result),
            self.binary_predicate_to_dblp(p.right, result)
            return

        # Simple predicate. The left member must be a attribute name of the entry.
        if not isinstance(p.left, str):
            raise RuntimeError("binary_predicate_to_dblp: left operand (%r) of %s must be a string" % (p.left, p))

        if p.left in ["author", "authors", "researcher", "conference"]:
            # String attribute. Only "CONTAINS" and "==" are supported. "==" behaves like "CONTAINS"
            if p.operator == operator.__eq__ or p.operator == operator.__contains__:
                result["prefix"] = self.get_dblp_name(p.right)
            else:
                raise RuntimeError("binary_predicate_to_dblp: unsupported operator (%s): %s" % (p.operator, p))
        else:
            # Other attributes. Only "==" is supported.
            if p.operator == operator.__eq__:
                result["suffix"] += "%20" + ("%s:%s" % (p.left, p.right))
            else:
                raise RuntimeError("binary_predicate_to_dblp: unsupported operator (%s): %s" % (p.operator, p))

    @staticmethod
    def to_doc_type(s :str) -> DocType:
        s = s.lower()
        if   s in {
            "article", "conference and workshop papers", "conference or workshop",
            "incollection", "inproceedings", "proceedings"
        }:
            return DocType.ARTICLE
        #elif s == "????":
        #    return DocType.COMM
        elif s == "journal articles":
            return DocType.JOURNAL
        elif s in {"informal publications", "reference works"}:
            return DocType.REPORT
        elif s in {"phdthesis", "books and theses", "parts in books or collections"}:
            return DocType.BOOKS_AND_THESES
        #elif s == "????":
        #    return DocType.HDR
        elif s == "editorship":
            return DocType.CHAPTER
        else:
            Log.warning("DblpConnector.to_doc_type: unknown type: %s" % s)
            return DocType.UNKNOWN

    def reshape_entry(self, query :Query, entry :dict) -> dict:
        if "type" in entry.keys():
            doc_type = DblpConnector.to_doc_type(entry["type"])
            entry["doc_type"]      = doc_type # Compatible with Hal "doc_type" values.
            entry["dblp_doc_type"] = doc_type # To compare Hal and DBLP doc_types.

        if len(query.attributes) > 0:
            keys = set(entry.keys()) & set(query.attributes)
            entry = {k : entry[k] for k in keys}

        for (k, v) in entry.items():
            if isinstance(v, str):
                try:
                    entry[k] = int(v)
                except ValueError:
                    pass

        if "authors" in entry.keys():
            if isinstance(entry["authors"], (dict, str)):
                entry["authors"] = [entry["authors"]]

            # Fix author names having homonyms are not well-named (e.g
            # "Giovanni Pau 0001" instead of "Giovanni Pau").
            entry["authors"] = [
                to_canonic_fullname(author).rstrip(" 0123456789") for author in entry["authors"]
            ]

            # Convert DBLP names to our names if needed.
            entry["authors"] = [
                self.m_map_rev_name.get(_to_canonic_fullname(author), author) \
                for author in entry["authors"]
            ]

        return entry

    def extract_entries(self, query :Query, results :list) -> list:
        entries = list()
        # If the query is not a standard Dblp object, we are pulling a bibliography
        # of a researcher who is identified by query.object (PID or fullname)
        is_bib_query = query.object not in {"publication", "researcher", "conference"}
        canonic_fullname = to_canonic_fullname(self.get_dblp_name(query.object)).rstrip("$") if is_bib_query else None
        try:
            raw_entries = results["result"]["hits"]["hit"]
            for raw_entry in raw_entries:
                entry               = raw_entry["info"]
                entry["dblp_id"]    = int(raw_entry["@id"])
                entry["dblp_score"] = int(raw_entry["@score"])

                try:
                    entry["authors"] = entry.pop("authors")["author"]
                except KeyError:
                    pass

                if is_bib_query:
                    # Unfortunately, DBLP seems unable to perform exact match search.
                    # For example searching "francois-durand$" returns "Jean-François
                    # Durand" publications.
                    # The following discards the false positives.
                    if isinstance(entry["authors"], str):
                        entry["authors"] = [entry["authors"]]

                    if canonic_fullname not in [
                        to_canonic_fullname(author) for author in entry["authors"]
                    ]:
                        #Log.warning("Ignoring %(title)s %(authors)s (homonym)" % entry)
                        continue

                entries.append(entry)
        except KeyError: # occurs if 0 DBLP publication found
            pass
        return entries

    def reshape_entries(self, query :Query, entries :list) -> list:
        return [self.reshape_entry(query, entry) for entry in entries]

    def query(self, query :Query) -> list:
        super().query(query)
        entries = list()
        if query.action == ACTION_READ:

            pid = None
            format = self.format
            object = ""
            url_options = list()
            if   query.object == "publication":
                object = "search/publ"
            elif query.object == "researcher":
                object = "search/author"
            elif query.object == "conference":
                object = "search/venue"
            else:
                fullname = query.object
                pid = self.map_dblp_id.get(fullname)
                dblp_name = self.get_dblp_name(fullname)
                object = "pid" if pid else "search/publ"
                # For the moment, DBLP only supports XML for pid-based queries.
                # https://dblp.org/pid/30/1446.xml
                if pid: format = "xml"
                else: url_options.append(dblp_name)

            if object == "pid":
                q_dblp = "%(server)s/%(object)s/%(pid)s.%(format)s" % {
                    "server" : self.api_url,
                    "object" : object,
                    "pid"    : pid,
                    "format" : format,
                }
            else:
                # WHERE
                if query.filters:
                    search = {
                        "prefix" : self.get_dblp_name(query.object),
                        "suffix" : ""
                    }
                    self.binary_predicate_to_dblp(query.filters, search)
                    url_options.append("%s%s" % (search["prefix"], search["suffix"]))

                # OFFSET and LIMIT
                url_options.append("h=%s" % query.limit if query.limit is not None else "h=9999")
                if query.offset:
                    url_options.append("f=%s" % query.offset)

                # Format of the result.
                url_options.append("format=%s" % format)
                q_dblp = "%(server)s/%(object)s/api?q=%(query)s" % {
                    "server" : self.api_url,
                    "object" : object,
                    "query"  : "&".join(url_options)
                }

            Log.info("--> DBLP: %s" % q_dblp)
            # http = urllib3.PoolManager(timeout=urllib3.Timeout(connect=1.0, read=2.0))
            # reply = http.request("GET", q_dblp, retries = 5)
            reply = requests.get(q_dblp)

            if reply.status_code == 200:
                data = reply.content.decode("utf-8")
                if format == "json":
                    result = json.loads(data)
                    entries = self.extract_entries(query, result)
                elif format == "xml":
                    data = data.replace("<i>", "")
                    data = data.replace("</i>", "")
                    result = xmltodict.parse(data, dict_constructor=dict)
                    # N.B. There are two other keys of interests
                    # - "co" : coauthors
                    # - "person" : information about the researcher

                    def xml_to_entry(d :dict) -> dict:
                        publication_type = next(iter(d.keys()))
                        entry = d[publication_type]
                        entry["type"] = publication_type
                        key = \
                            "author" if "author" in entry.keys() else \
                            "editor" if "editor" in entry.keys() else \
                            None
                        if key:
                            # Sometimes, author is represented by a dict with key
                            # '@orcid' and '#text'
                            if isinstance(entry[key], str):
                                entry[key] = [entry[key]]

                            entry["authors"] = [
                                author["#text"] if isinstance(author, dict) else author \
                                for author in entry[key]
                            ]
                        else:
                            Log.warning("No author found for this DBLP publication:\n%s" % \
                                pformat(entry)
                            )
                        return entry

                    entries = [xml_to_entry(d) for d in result["dblpperson"]["r"]]
                else:
                    raise RuntimeError("Format not implemented: %s" % self.format)
            else:
                raise RuntimeError("Cannot get reply from %s" % self.api_url)

        return self.answer(query, self.reshape_entries(query, entries))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import re, sys
from html       import escape
from .connector import Connector
from .query     import Query

PATTERN_URL = re.compile("http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+")

#def str_to_html(s :str) -> str:
#    if PATTERN_URL.match(s):
#        return "<a href=%s>%s</a>" % (s, s)
#    else:
#        return s.rstrip(".").rstrip(",")

def str_to_html(s :str) -> str:
    #return escape(s).rstrip(".").rstrip(",")
    return s.rstrip(".").rstrip(",")

def to_html_link(d :dict, k :str, v :str) -> str:
    try:
        return "<a href=\"%s\">%s</a>" % (d[k], escape(v))
    except KeyError:
        return v

def value_to_html(x) -> str:
    if isinstance(x, dict):
        ret = dict_to_html(x, x.keys())
    elif isinstance(x, str):
        if PATTERN_URL.match(x):
            ret = "<a href=\"%s\">%s</a>" % (x, escape(x))
        else:
            ret = str_to_html(x)
    elif isinstance(x, (list, set)):
        ret = ", ".join([value_to_html(value) for value in x])
    else:
        ret = escape("%s" % x)
    return ret

def html(s :str):
    """
    Evaluate HTML code in a Jupyter Notebook.
    Args:
        s: A str containing HTML code.
    """
    from IPython.display import display, HTML
    chart = HTML(s)
    # or chart = charts.plot(...)
    display(chart)

def print_error(x):
    """
    Print an error in a Jupyter Notebook.
    """
    print(str(x), file = sys.stderr)

def entries_to_html(entries :list, map_attribute_label :dict = dict(), attributes :list = None) -> str:
    """
    Export to HTML a list of dict.
    Args:
        entries: A list of dicts
        map_attribute_label: A dict {str : str} which maps each entry key with the column header
            to display.
        attributes: The subset of keys to display.
    """

def entries_to_html(
    entries :list,
    map_attribute_label :dict = None,
    attributes :list = None,
    keep_entry_if :callable = None
) -> str:
    if not attributes:
        attributes = list(entries[0].keys())
    if not keep_entry_if:
        keep_entry_if = lambda i, entry: True
    if not map_attribute_label:
        map_attribute_label = dict()
    return """
        <div style="width:100%%; overflow-x: auto;white-space: nowrap;">
            <table>
                %(header)s
                %(rows)s
            </table>
        </div>
        """ % {
            "header" : "<tr>%s</tr>" % "".join(
                ["<th>Index</th>"] + [
                    "<th style='text-align: left;white-space: nowrap;'>%s</th>" % str(map_attribute_label.get(attribute, attribute))
                    for attribute in attributes
                ]
            ),
            "rows" : "".join([
                """
                <tr>
                    <td>%(index)d</td>
                    %(values)s
                </tr>
                """ % {
                    "index" : i,
                    "values" : "".join([
                        "<td style='text-align: left;white-space: nowrap;'>%s</td>" % value_to_html(entry.get(attribute))
                        for attribute in attributes
                    ])
                }
                for (i, entry) in enumerate(entries)
                if keep_entry_if(i, entries)
            ]),
        }

def connector_to_html(connector :Connector, **kwargs) -> str:
    attributes = kwargs.get("attributes")
    if not attributes:
        attributes = connector.attributes(None)
    entries = connector.query(Query(**kwargs))
    return entries_to_html(entries, attributes=attributes)

def entry_to_html(entry :dict, map_attribute_label :dict = dict(), attributes :list = None) -> str:
    return entries_to_html([entry], map_attribute_label, attributes)

# OBSOLETE
def dict_to_html(d :dict, attributes :list, map_attribute_label :dict = dict()) -> str:
    return entry_to_html(d, map_attribute_label, attributes)

#----------------------------------------------------------------------------------
# Extensions depending on BeautifulSoup
#----------------------------------------------------------------------------------

try:
    # See https://matix.io/extract-text-from-webpage-using-beautifulsoup-and-python/
    from bs4 import BeautifulSoup

    HTML_BLACKLIST_TAGS = {
        "[document]",
        "noscript",
        "header",
        "html",
        "head",
        "link",
        "input",
        "meta",
        "script",
        "style"
    }

    def html_to_text(s_html :str, blacklist :set = HTML_BLACKLIST_TAGS) -> str:
        """
        Convert an HTML page to text, by discarding javascript and css related
        to the site.
        Args:
            s_html: A str containing HTML.
            blacklist: A set of str (lowercase) corresponding to HTML tags
                that must be ignored.
        """
        soup = BeautifulSoup(s_html, features = "lxml")
        soup.find_all(s_html, "html.parser")
        text = soup.find_all(text = True)
        l = list()
        for t in text:
            s = t.strip()
            if s and t.parent.name not in blacklist:
                #print("<%s>: %r" % (t.parent.name, s))
                s = re.sub("<[^>]*>", "", s)
                l += [s]
        s = " ".join(l)
        return re.sub(r"[\n]+", "\n\n", s)

    HTML_MAP_WHITELIST_ATTRS = {
        "a"   : ["href"],
        "img" : ["src"],
    }

    # Based on https://gist.github.com/revotu/21d52bd20a073546983985ba3bf55deb
    def remove_all_attrs_except_saving(soup, whitelist = HTML_MAP_WHITELIST_ATTRS):
        """
        Remove all attributes except some.
        Args:
            soup: A BeautifulSoup instance, modified in place.
            whitelist: A dict {tag : list(attr)} where tag is an HTML tag and attr
                an HTML attribute.
        """
        for tag in soup.find_all(True):
            if tag.name not in whitelist.keys():
                tag.attrs = {}
            else:
                attrs = dict(tag.attrs)
                allowed_attrs = whitelist[tag.name]
                for attr in attrs:
                    if attr not in allowed_attrs:
                        del tag.attrs[attr]
                    elif attr == "href" and not attr.startswith("http"):
                        tag.attrs[attr] = "#"
        return soup

    def remove_tags(soup, blacklist :set = HTML_BLACKLIST_TAGS):
        """
        Remove some HTML tags.
        Args:
            soup: A BeautifulSoup instance, modified in place.
            blacklist: A list of str, where each str is an HTML tag.
        """
        for tag in blacklist:
            while True:
                attr = getattr(soup, tag)
                if attr:
                    attr.decompose()
                else:
                    break

    def sanitize_html(
        s_html       :str,
        blacklist    :set  = HTML_BLACKLIST_TAGS,
        remove_attrs :bool = True
    ) -> str:
        """
        Remove from an HTML string irrelevants HTML blocks and attributes.

        /!\ This function is SLOW so do not use it on large corpus!

        Args:
            s_html: A str instance containing HTML.
            blacklist: List of blacklisted HTML tags.
        Returns:
            The sanitized string.
        """
        soup = BeautifulSoup(s_html, features = "lxml")
        if remove_attrs:
            remove_all_attrs_except_saving(soup)
        body = soup.find("body")
        if not body:
            return None
        remove_tags(body, blacklist)
        return str(body.prettify())

except ImportError as e:
    #print(
    #    "\n  ".join([
    #        "Please install bs4",
    #        "APT: sudo apt install python3-bs4",
    #        "PIP: sudo pip3 install bs4",
    #    ]),
    #    file = sys.stderr
    #)
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import codecs, json, os, traceback
from minifold.entries_connector import EntriesConnector

def identity(json_py) -> list:
    # Minifold expects a list of dict, so if you are looking
    # for a specific part of the json file, this callback
    # is an opportunity to extract the data of interests.
    return json_py

def load_json_from_str(json_data :str, extract_json = identity) -> list:
    assert isinstance(json_data, str)
    entries = extract_json(json.loads(json_data))
    assert isinstance(entries, list), "Minifold expects a list of dict, got %s" % type(entries)
    return entries

def load_json_from_file(json_filename :str, extract_json = identity) -> list:
    entries = None
    with codecs.open(json_filename, encoding = "utf8") as f_json:
        entries = load_json_from_str(f_json.read(), extract_json)
    return entries

class JsonConnector(EntriesConnector):
    def __init__(self, json_data :str, extract_json = identity):
        entries = load_json_from_str(json_data, extract_json)
        assert isinstance(entries, list)
        super().__init__(entries)

class JsonFileConnector(JsonConnector):
    def __init__(self, json_filename :str, extract_json = identity):
        with codecs.open(json_filename, encoding = "utf8") as f_json:
            super().__init__(f_json.read(), extract_json)
    # For the moment we only support ACTION_READ queries,
    # but this could be extended



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"


def is_multiple_key(key) -> bool:
    return isinstance(key, (set, frozenset, list))

def closure(key, fds) -> set:
    ret = (key | {k for k in key}) if is_multiple_key(key) else {key}
    changed = True

    i = 0
    while changed == True:
        changed = False
        i += 1
        for (key, column) in fds:
            if key in ret and column not in ret:
                ret.add(column)
                changed = True

    if is_multiple_key(key):
        ret.remove(key)

    return ret


def minimal_cover(fds :set) -> list:
    min_fds = fds.copy()

    for (key, column) in fds:
        # Try to remove (key, column) and see if we cover the closure remains the same
        if column in closure(key, min_fds - {(key, column)}):
            min_fds -= {(key, column)}

    for (key, column) in min_fds.copy():
        if is_multiple_key(key):
            # Check whether each k attribute is required, otherwise, discard it.
            for k in key:
                if k in closure(key - k, min_fds - {(key, column)}):
                    min_fds -= {(key, column)}
                    min_fds |= {(key - k, column)}

    return min_fds

try:
    from pybgl.graph import DirectedGraph, EdgeDescriptor, edges, source, target

    def edge_to_pair(e :EdgeDescriptor, g :DirectedGraph) -> tuple:
        return (source(e, g), target(e, g))

    def minimal_cover_graph(g :DirectedGraph) -> set:
        return minimal_cover({edge_to_pair(e, g) for e in edges(g)})
except ImportError:
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import datetime, functools, json, os, pickle, sys, traceback

from functools              import partial
from pprint                 import pprint

from minifold.connector     import Connector
from minifold.filesystem    import check_writable_directory, mtime, mkdir, rm
from minifold.query         import Query, ACTION_READ
from minifold.log           import Log

# TODO:
# For the moment the cache is class-name based. It should rather
# be identify by the connector setup and the underlying connectors
# (if any)

# TODO:
# For the moment, the cache is only used if the exact same Query
# was issued over the past. But we should be able to reuse the cache
# if a less strict query has been issued, and use
# Connector.reshape_entries afterwards.

class CacheConnector(Connector):
    def __init__(self, child):
        self.child = child

    def attributes(self, object :str) -> set:
        return self.child.attributes(object)

    def callback_read(self, query): # raises RuntimeError
        raise RuntimeError("Must be overloaded")

    def callback_write(self, query, data): # raises RuntimeError
        raise RuntimeError("Must be overloaded")

    def clear_query(self, query :Query):
        pass

    def clear_cache(self):
        pass

    def read(self, query :Query):
        (data, success) = (None, False)
        try:
            data = self.callback_read(query)
            success = (data != None)
        except:
            Log.error(
                "CacheConnector.read(%s): Cannot read cache:\n%s" % (
                    query,
                    traceback.format_exc()
                )
            )
        return (data, success)

    def write(self, query :Query, data) -> bool:
        success = True
        try:
            self.callback_write(query, data)
        except:
            Log.error(
                "CacheConnector.write(%s, %s): Cannot write cache:\n%s" % (
                    traceback.format_exc(),
                    query,
                    data
                )
            )
            success = False
        return success

    def is_cached(self, query :Query) -> bool:
        raise RuntimeError("Must be overloaded")

    def is_cachable(self, query :Query, data) -> bool:
        return True

    def query(self, query :Query):
        (data, success) = (None, False)
        if self.is_cached(query):
            (data, success) = self.read(query)
            if not success:
                Log.warning("CacheConnector.query(%s): Unreadable cache" % query)
        if not success:
            data = self.child.query(query)
        if query.action == ACTION_READ and self.is_cachable(query, data):
            self.write(query, data)
        return self.answer(query, data)

# Default parameters, used to initialize StorageCacheConnector class members.
DEFAULT_CACHE_STORAGE_BASE_DIR = os.path.join(os.path.expanduser("~"), ".minifold", "cache")
DEFAULT_CACHE_STORAGE_LIFETIME = datetime.timedelta(days = 3)

def make_cache_dir(base_dir :str, sub_dir :str = ""):
    return os.path.join(base_dir, sub_dir) if sub_dir else base_dir

class StorageCacheConnector(CacheConnector):
    base_dir = DEFAULT_CACHE_STORAGE_BASE_DIR
    lifetime = DEFAULT_CACHE_STORAGE_LIFETIME

    def __init__(
        self,
        child :Connector,
        callback_load = None,
        callback_dump = None,
        lifetime   :datetime.timedelta = None,
        cache_dir  :str = None,
        read_mode  :str = "r",
        write_mode :str = "w",
        extension  :str = ""
    ):
        super().__init__(child)
        self.callback_load = callback_load
        self.callback_dump = callback_dump
        self.lifetime   = lifetime if lifetime is not None else StorageCacheConnector.lifetime
        self.cache_dir  = cache_dir if cache_dir else \
                          make_cache_dir(StorageCacheConnector.base_dir, child.__class__.__name__)
        self.read_mode  = read_mode
        self.write_mode = write_mode
        self.extension  = extension

    def make_cache_filename(self, query :Query) -> str:
        return os.path.join(self.cache_dir, str(query) + self.extension)

    def clear_query(self, query :Query):
        cache_filename = self.make_cache_filename(query)
        if os.path.exists(cache_filename):
            Log.debug("StorageCacheConnector: Removing query [%s]" % cache_filename)
            rm(cache_filename)

    def clear_cache(self):
        if os.path.exists(self.cache_dir) and os.path.isdir(self.cache_dir):
            Log.debug("StorageCacheConnector: Removing cache [%s]" % self.cache_dir)
            rm(self.cache_dir, recursive=True)

    @staticmethod
    def is_fresh_cache(cache_filename :str, lifetime :datetime.timedelta) -> bool:
        is_fresh = True
        if lifetime:
            t_now = datetime.datetime.utcnow()
            t_cache = mtime(cache_filename)
            is_fresh = (t_now - t_cache) < lifetime
            #Log.debug("t_now(%s) - t_cache(%s) = %s ?< lifetime %s" % (
            #    t_now, t_cache, (t_now - t_cache), lifetime
            #))
        return is_fresh

    def is_cached(self, query :Query) -> bool:
        ret = False
        cache_filename = self.make_cache_filename(query)
        if os.path.exists(cache_filename):
            ret = StorageCacheConnector.is_fresh_cache(cache_filename, self.lifetime)
        return ret

    def callback_read(self, query :Query) -> tuple:
        data = None
        cache_filename = self.make_cache_filename(query)
        if self.is_cached(query):
            with open(cache_filename, self.read_mode) as f:
                data = self.callback_load(f)
            Log.debug("Cache hit: [%s]" % cache_filename)
        return data

    def callback_write(self, query :Query, data):
        cache_filename = self.make_cache_filename(query)
        directory = os.path.dirname(cache_filename)
        mkdir(directory)
        check_writable_directory(directory)
        with open(cache_filename, self.write_mode) as f:
            self.callback_dump(data, f)

class PickleCacheConnector(StorageCacheConnector):
    def __init__(
        self,
        child     :Connector,
        lifetime  :datetime.timedelta = StorageCacheConnector.lifetime,
        cache_dir :str                = None
    ):
        super().__init__(
            child,
            pickle.load, pickle.dump,
            lifetime, cache_dir,
            "rb", "wb", ".pkl"
        )

class JsonCacheConnector(StorageCacheConnector):
    def __init__(
        self,
        child     :Connector,
        lifetime  :datetime.timedelta = StorageCacheConnector.lifetime,
        cache_dir :str                = None
    ):
        super().__init__(
            child,
            json.load, partial(json.dump, indent=4),
            lifetime, cache_dir,
            "r", "w", ".json"
        )





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob, Fabien Mathieu"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com, fabien.mathieu@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import requests
from .log       import Log
from .singleton import Singleton

class Proxy(dict, metaclass=Singleton):
    pass

def proxy_enable(host :str, port :int, protocols :list = ["http", "https"]):
    proxy = Proxy()
    for protocol in protocols:
        url = "%s://%s:%s" % (protocol, host, port)
        proxy[protocol] = url

def proxy_disable():
    proxy = Proxy()
    proxy.clear()

def make_session() -> requests.Session:
    session = requests.Session()
    proxy = Proxy()
    if proxy:
        Log.info("Setting proxy = %s" % proxy)
        session.proxies.update(proxy)
    return session

#---------------------------------------------------------------
# Example to use a local proxy
#---------------------------------------------------------------

PROXY_LOCALHOST_IP    = "127.0.0.1"
PROXY_LOCALHOST_PORT  = 8080

def proxy_enable_localhost():
    proxy_enable(PROXY_LOCALHOST_IP, PROXY_LOCALHOST_PORT)



#!/usr/bin/env python3
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from html.parser                    import HTMLParser
from html.entities                  import name2codepoint

from minifold.entries_connector     import EntriesConnector

class HtmlTableParser(HTMLParser):
    def __init__(self, columns :list, output_list :list, keep_entry = None):
        """
        Constructor.
        Args:
            columns: list of string mapping the attribute name
                corresponding with the index. If data is fetch
                for columns having a greater index than
                len(columns), columns[-1] is used, and this
                key may store a list of string values instead
                of a single string. This allow to store data
                stored among several columns in a single attribute.
            output_list: reference to an output list where the
                data will be outputed (one dict per row, one
                key/value per column).
            keep_entry: callback which determine whether an
                must entry must be kept or discard. Pass None
                to filter nothing. This is the opportunity to
                discard a header or irrelevant row.
        """
        HTMLParser.__init__(self)
        self.fetch_data = False
        self.columns    = columns
        self.index      = 0
        self.entries    = output_list
        self.entry      = dict()
        self.value      = str()
        self.keep_entry = keep_entry

    def attributes(self, object :str) -> set:
        return set(self.columns)

    def handle_starttag(self, tag, attrs):
        if tag == "td":
            # Enable fetch data
            self.fetch_data = True

    def handle_endtag(self, tag):
        if tag == "td": # Push key/value
            # Disable fetch data
            self.fetch_data = False

            # Push new key/value pair
            key = self.columns[self.index] if self.index < len(self.columns) else self.columns[-1]
            if key in self.entry.keys():
                current_value = self.entry[key]
                if not isinstance(current_value, list):
                    self.entry[key] = [current_value]
                if self.value:
                    self.entry[key].append(self.value)
            else:
                if self.value:
                    self.entry[key] = self.value

            # Reset key/value pair
            self.value = str()
            self.index += 1
        elif tag == "tr":
            # Push entry
            if self.keep_entry == None or self.keep_entry(self.entry):
                self.entries.append(self.entry)

            # Reset entry
            self.index = 0
            self.entry = dict()

    def handle_data(self, data):
        data = data.strip()
        if self.fetch_data == True and data:
            self.value += data

def html_table(filename :str, columns :list, keep_entry = None) -> list:
    entries = list()
    parser = HtmlTableParser(columns, self.m_entries, keep_entry)
    with open(filename, "r") as f:
        s = f.read()
        parser.feed(s)
    return entries

class HtmlTableConnector(EntriesConnector):
    def __init__(self, filename :str, columns :list, keep_entry = None):
        """
        Constructor.
        Args:
            filename: Input HTML filename.
            columns: list of string mapping the attribute name
                corresponding with the index. If data is fetch
                for columns having a greater index than
                len(columns), columns[-1] is used, and this
                key may store a list of string values instead
                of a single string. This allow to store data
                stored among several columns in a single attribute.
            output_list: reference to an output list where the
                data will be outputed (one dict per row, one
                key/value per column).
            keep_entry: callback which determine whether an
                must entry must be kept or discard. Pass None
                to filter nothing. This is the opportunity to
                discard a header or irrelevant row.
        """
        self.m_entries = list()
        self.m_parser = HtmlTableParser(columns, self.m_entries, keep_entry)
        with open(filename, "r") as f:
            s = f.read()
            self.m_parser.feed(s)
        super().__init__(self.m_entries)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector  import Connector
from .query      import Query, ACTION_READ, action_to_str

class EntriesConnector(Connector):
    def __init__(self, entries :list):
        super().__init__()
        self.m_keys = set()
        if len(entries) > 0:
            for entry in entries:
                self.m_keys |= set(entry.keys())
        self.m_entries = entries

    def attributes(self, obj :str = None) -> set:
        return set(self.m_keys)

    def query(self, query :Query) -> list:
        super().query(query)
        ret = list()
        if query.action == ACTION_READ:
            ret = self.reshape_entries(query, self.entries)
        else:
            raise RuntimeError("EntriesConnector::query: %s not yet implemented" % action_to_str(query.action))
        return self.answer(query, ret)

    @property
    def entries(self) -> list:
        return self.m_entries





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

#---------------------------------------------------------------------------------
# Requests
#---------------------------------------------------------------------------------

import sys

if sys.version_info < (3, 5): # async, await
    raise RuntimeError("asyncio requires python>=3.5")

try:
    import requests
except ImportError as e:
    print(
        "\n".join([
            "Please install requests",
            "  APT: sudo apt install python3-requests",
            "  PIP: sudo pip3 install --upgrade requests",
        ]),
        file = sys.stderr
    )
    raise e

import asyncio, datetime, re
import concurrent.futures

from copy              import deepcopy
from functools         import partial
from .connector        import Connector
from .query            import ACTION_READ, Query
from .html             import html_to_text
from .log              import Log
from .proxy            import make_session
from .request_cache    import install_cache

DEFAULT_TIMEOUT = (1.0, 2.0) # (connect timeout, read timeout)
#DEFAULT_TIMEOUT = (0.5, 1.0) # (connect timeout, read timeout)

def now() -> str:
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

def download(url :str, timeout = DEFAULT_TIMEOUT, cache_filename = None):
    """
    Download the content related to a given URL.
    Args:
        url: A str containing the target URL.
        timeout: A tuple (float, float) corresponding to the
            (connect timeout, read timeout).
        cache_filename: A str containing path to the cache to use.
            Pass None to use the default cache.
    Raises:
        requests.exceptions.ConnectionError
        requests.exceptions.ConnectTimeout
        requests.exceptions.ContentDecodingError
        requests.exceptions.ReadTimeout
        requests.exceptions.SSLError
    Returns:
        The corresponding response.
    """
    install_cache(cache_filename)
    try:
        #Log.debug("download: GET %s (timeout = %s)" % (url, timeout))
        # https://stackoverflow.com/questions/8287628/proxies-with-python-requests-module
        session = make_session()
        return session.get(
            url,
            timeout = timeout
        )
    except Exception as e:
        Log.warning("download: GET %s (timeout = %s): %s" % (url, timeout, e))
        return e

def trim_http(url :str) -> str:
    """
    Discard the "http://" or "https://" prefix from an url.
    Args:
        url: A str containing an URL.
    Returns:
        The str corresponding to the trimmed URL.
    """
    return re.sub("https?://", "", url, flags = re.IGNORECASE)

# https://skipperkongen.dk/2016/09/09/easy-parallel-http-requests-with-python-and-asyncio/
async def downloads_async(
    urls :list,
    timeout = DEFAULT_TIMEOUT,
    return_exceptions :bool = True
):
    """
    Asynchronous download procedure.
    Args:
        urls: An iterable over strings, each of them corresponding to an URL.
        timeout: A tuple (float, float) corresponding to the
            (connect timeout, read timeout).
    Returns:
        A dict({str : ?}} mapping for each queried URL the corresponding
        contents (if successful), the corresponding Exception otherwise.
        See download() documentation for possible exceptions.
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=500) as executor:
        loop = asyncio.get_event_loop()
        futures = [
            loop.run_in_executor(
                executor,
                partial(download, timeout=timeout, cache_filename=None),
                url
            ) for url in urls
        ]
        return {
            url : response
            for (url, response) in zip(
                urls,
                await asyncio.gather(*futures, return_exceptions = return_exceptions)
            )
        }

def downloads(*args) -> dict:
    """
    Perform multiple downloads in parallel. See downloads_async for further details.
    """
    Log.debug("[%s] downloads: start" % now())
    # https://stackoverflow.com/questions/46727787/runtimeerror-there-is-no-current-event-loop-in-thread-in-async-apscheduler
    #loop = asyncio.get_event_loop()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    ret = loop.run_until_complete(downloads_async(*args))
    Log.debug("[%s] downloads: %d URLs fetched" % (now(), len(ret)))
    return ret

#---------------------------------------------------------------------------------
# Minifold
#---------------------------------------------------------------------------------

def extract_response(response, extract_text :bool = True):
    """
    Extract from a response the corresponding contents or Exception.
    Args:
        response: A requests.Response or an Exception instance.
        extract_text: A bool indicating if text must be extracted
            from HTML.
    Returns:
        The corresponding Exception or str.
    """
    if isinstance(response, Exception):
        return response # Forward exception
    text = response.text
    return html_to_text(text) if extract_text else raw

RE_VALID_URL = re.compile("https?://.*")

class DownloadConnector(Connector):
    def __init__(
        self,
        map_url_out :dict,
        child :Connector,
        downloads = downloads,
        extract_response = extract_response
        # TODO Pass downloads() parameters
    ):
        """
        Constructor.
        Args:
            map_url_out: dict(str : str) entry attribute containing an URL to another
                entry attribute that will store the contents provided by this URL.
                Example: {"url" : "html_content"}
            child: A child Connector.
            downloads: Callback(urls) -> dict(url : content) where
                    urls is an iterable of URLs and where.
                    the returned dict maps each urls and the corresponding response.
                Note: You could pass partial(download, ...) to customize timouts.
            extract_response: Callback(response) -> str
        """
        self.map_url_out = map_url_out
        self.child = child
        self.downloads = downloads
        self.extract_response = extract_response

    def query(self, query :Query) -> list:
        """
        Handle an input Query.
        Args:
            query: The input Query.
        Returns:
            The corresponding result.
        """
        from pprint import pformat
        if query.action == ACTION_READ:
            # Pull child records
            q_child = deepcopy(query) # TODO: rework q_child like in LambdasConnector
            q_child.attributes = set(query.attributes) - set(self.map_url_out.values())
            entries = self.child.query(q_child)

            attributes = set(query.attributes) & set(self.map_url_out.values()) \
                if query.attributes else set(self.map_url_out.values())

            # If some URLs must be queried
            if attributes:
                # Fetch URLs
                urls = set()
                for attr_url in self.map_url_out.keys():
                    urls |= {
                        entry.get(attr_url) for entry in entries
                        if isinstance(entry.get(attr_url), str) \
                           and RE_VALID_URL.match(entry.get(attr_url))
                    }
                Log.debug("Starting fetching %d URLs" % len(urls))
                map_url_response = self.downloads(urls)
                if self.extract_response:
                    map_url_response = {
                        url : self.extract_response(response)
                        for (url, response) in map_url_response.items()
                    }

                # Dispatch responses in appropriate entry attributes
                for (attr_url, attr_out) in self.map_url_out.items():
                    for entry in entries:
                        url = entry.get(attr_url)
                        if url:
                            entry[attr_out] = map_url_response[url]

            # Forward to parent Connector once entry are reshaped.
            entries = self.reshape_entries(query, entries)
            return entries
        else:
            raise RuntimeError("Action not implemented: %s" % query.action)

    def attributes(self, object :str) -> set:
        return set(self.child.attributes(object)) | set(self.map_url_out.values())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import os, sys
from minifold.cache         import DEFAULT_CACHE_STORAGE_BASE_DIR, DEFAULT_CACHE_STORAGE_LIFETIME
from minifold.filesystem    import check_writable_directory, mkdir

try:
    import requests_cache
except ImportError as e:
    print(
        "Please install requests-cache.\n" + \
        "  APT: sudo apt install python3-requests-cache\n" + \
        "  PIP: sudo pip3 install --upgrade requests-cache\n",
        file = sys.stderr
    )
    raise e

def install_cache(cache_filename :str = None):
    if not cache_filename:
        directory = DEFAULT_CACHE_STORAGE_BASE_DIR
        cache_filename = os.path.join(DEFAULT_CACHE_STORAGE_BASE_DIR, "requests_cache")
    else:
        directory = os.path.dirname(cache_filename)
    mkdir(directory)
    check_writable_directory(directory)
    requests_cache.install_cache(
        cache_filename,
        expire_after = DEFAULT_CACHE_STORAGE_LIFETIME
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .query                 import Query, ACTION_READ

# The following schema summarizes the different joins
# http://4.bp.blogspot.com/-_HsHikmChBI/VmQGJjLKgyI/AAAAAAAAEPw/JaLnV0bsbEo/s1600/sql%2Bjoins%2Bguide%2Band%2Bsyntax.jpg

INNER_JOIN      = 1 # Returns matching records
LEFT_JOIN       = 2 # Returns left entries + matching right entries (=> right attributes may be None)
RIGHT_JOIN      = 3 # Returns right entries + matching left entries (=> left attributes may be None)
FULL_OUTER_JOIN = 4 # Returns left and right entries while merging matching entries

def merge_dict(l :dict, r :dict) -> dict:
    ret = l.copy()
    ret.update(r)
    return ret

def join_mode_to_string(mode :int) -> str:
    if   mode == INNER_JOIN:      return "INNER JOIN"
    elif mode == LEFT_JOIN:       return "LEFT JOIN"
    elif mode == RIGHT_JOIN:      return "RIGHT JOIN"
    elif mode == FULL_OUTER_JOIN: return "FULL OUTER JOIN"
    return "?"

def are_joined_if(l :dict, r :dict, f) -> bool:
    try:
        return f(l, r)
    except KeyError as e:
        raise RuntimeError("Cannot join\n\tl = %r\n\tr = %r\n%s" % (l, r, e))

def inner_join_if(l_entries :list, r_entries :list, f, match_once = True, merge = merge_dict) -> list:
    ret = list()
    for l in l_entries:
        for r in r_entries:
            if are_joined_if(l, r, f):
                ret.append(merge(l, r))
                if match_once: break
    return ret

def left_join_if(l_entries :list, r_entries :list, f, match_once = True, merge = merge_dict) -> list:
    ret = list()
    if len(r_entries) > 0:
        for l in l_entries:
            joined = False
            for r in r_entries:
                if are_joined_if(l, r, f):
                    joined = True
                    ret.append(merge(l, r))
                    if match_once: break
            if joined == False:
                entry = l
                missing_keys = set(r.keys()) - set(l.keys())
                for k in missing_keys:
                    entry[k] = None
                ret.append(entry)
    else:
        ret = l_entries
    return ret

def right_join_if(l_entries :list, r_entries :list, f, match_once = True) -> list:
    return left_join_if(r_entries, l_entries, lambda l,r: f(r,l), match_once)

def full_outer_join_if(l_entries :list, r_entries :list, f, match_once = True) -> list:
    # Get every left entries (joinable or not)
    ret = left_join_if(l_entries, r_entries, f, match_once)

    # Retrieve left keys
    l_keys = set()
    for l in l_entries:
        l_keys |= {k for k in l.keys()}

    # Get missing right entries
    for r in r_entries:
        joined = False
        for l in l_entries:
            if are_joined_if(l, r, f):
                joined = True
                break

        if joined == False:
            entry = r
            missing_keys = set(l_keys) - set(r.keys())
            for k in missing_keys:
                entry[k] = None
            ret.append(entry)

    return ret

class JoinIfConnector(Connector):
    def __init__(self, left :Connector, right :Connector, join_if, mode :int = INNER_JOIN):
        super().__init__()
        self.m_left = left
        self.m_right = right
        self.m_join_if = join_if
        self.m_left_entries = list()
        self.m_right_entries = list()
        self.m_mode = mode

    def attributes(self, object :str):
        return self.m_left.attributes(object) \
             | self.m_right.attributes(object)

    def query(self, query :Query) -> list:
        super().query(query)
        self.m_left_entries.clear()
        self.m_right_entries.clear()
        self.m_left_entries = self.left.query(query)
        self.m_right_entries = self.right.query(query)
        if   self.m_mode == INNER_JOIN:
            entries = inner_join_if(self.m_left_entries, self.m_right_entries, self.m_join_if)
        elif self.m_mode == LEFT_JOIN:
            entries = left_join_if(self.m_left_entries, self.m_right_entries, self.m_join_if)
        elif self.m_mode == RIGHT_JOIN:
            entries = right_join_if(self.m_left_entries, self.m_right_entries, self.m_join_if)
        elif self.m_mode == FULL_OUTER_JOIN:
            entries = full_outer_join_if(self.m_left_entries, self.m_right_entries, self.m_join_if)
        else:
            raise ValueError("JoinIfConnector::answer: Invalid mode %s:" % self.m_mode)
        return self.answer(query, entries)

    @property
    def mode(self):
        return self.m_mode

    @property
    def left(self):
        return self.m_left

    @property
    def right(self):
        return self.m_right

    @property
    def join_if(self):
        return self.m_join_if




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .query                 import Query

def limit(entries :list, lim :int) -> list:
    return entries[:lim]

class LimitConnector(Connector):
    def __init__(self, child, lim :int):
        super().__init__()
        self.m_child = child
        self.m_lim = lim

    @property
    def child(self):
        return self.m_child

    def attributes(self, object :str) -> set:
        return self.child.attributes(object)

    @property
    def limit(self):
        return self.m_lim

    def query(self, q :Query) -> list:
        super().query(q)
        return self.answer(
            q,
            limit(
                self.m_child.query(q),
                self.m_lim
            )
        )




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"


def in_ipynb() -> bool:
    """
    Tests whether the code is running inside a Jupyter Notebook.
    Returns:
        True iff the code is running inside a Jupyter Notebook.
    """
    try:
        return str(type(get_ipython())) == "<class 'ipykernel.zmqshell.ZMQInteractiveShell'>"
    except NameError:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import operator
from .dict_util import reverse_dict

def __in__(x, y) -> bool:
    return operator.__contains__(y, x)

OPERATORS = {
    ">"   : operator.__gt__,
    "<"   : operator.__lt__,
    ">="  : operator.__ge__,
    "<="  : operator.__le__,
    "=="  : operator.__eq__,
    "!="  : operator.__ne__,
    "||"  : operator.__or__,
    "OR"  : operator.__or__,
    "&&"  : operator.__and__,
    "AND" : operator.__and__,
    "^"   : operator.__xor__,
    "XOR" : operator.__xor__,
    "IN"  : __in__,
    "CONTAINS" : operator.__contains__,
}

OPERATORS_TO_STR = reverse_dict(OPERATORS)

class BinaryPredicate:
    def __init__(self, left, operator, right):
        self.m_left     = left
        self.m_operator = OPERATORS.get(operator, operator)
        self.m_right    = right
        assert not isinstance(self.m_operator, str)

    @property
    def left(self):
        return self.m_left

    @property
    def operator(self):
        return self.m_operator

    @property
    def right(self):
        return self.m_right

    def __str__(self) -> str:
        return "%s %s %s" % (self.left, OPERATORS_TO_STR[self.operator], self.right)

    def match(self, entry :dict) -> bool:
        try:
            if self.operator in [operator.__or__, operator.__and__, operator.__xor__]:
                return self.operator(self.left(entry), self.right(entry))
            else:
                left = entry[self.left]
                right = self.right
                return self.operator(left, right)
        except KeyError:
            return (self.operator == operator.__eq__ and self.right == None)

    def __call__(self, entry :dict) -> bool:
        return self.match(entry)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .query                 import Query, ACTION_READ

def select(entries :list, attributes :list) -> list:
    return [{k : entry[k] for k in attributes} for entry in entries]

class SelectConnector(Connector):
    def __init__(self, child, attributes):
        super().__init__()
        self.m_child = child
        self.m_attributes = attributes

    def attributes(self, object :str) -> set:
        return set(self.m_attributes) & self.m_child.attributes(object)

    @property
    def child(self):
        return self.m_child

    def query(self, query :Query) -> list:
        super().query(query)
        q = query.copy()
        q.attributes = [attribute for attribute in q.attributes if attribute in self.m_attributes]
        return self.answer(
            query,
            select(
                self.m_child.query(q),
                self.m_attributes
            )
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Dylan Sellar"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from minifold.binary_predicate    import BinaryPredicate
from minifold.connector           import Connector
from minifold.query               import Query, ACTION_READ
from minifold.where               import where
from pymongo                      import MongoClient

class MongoConnector(Connector):
    def __init__(self, mongo_url :str, db_name :str):
        self.client = self.connect(mongo_url)
        self.db_name = db_name
        self.use_database(db_name)

    def connect(self, mongo_url :str) -> MongoClient:
        return MongoClient(mongo_url)

    def use_database(self, db_name :str):
        self.db = self.client[db_name]

    def attributes(self, obj :str = None) -> set:
        return set()

    def query(self, query :Query) -> list:
        super().query(query)
        ret = list()
        assert self.db
        if query.action == ACTION_READ:
            ret = list(
                self.db[query.object].find(
                    {},
                    {k : 1 for attr in query.attributes} if query.attributes else None
                )
                    .limit(query.limit if query.limit else 0)
                    .skip(query.offset if query.offset else 0)
            )
            if not query.filters:
                pass
            elif isinstance(query.filters, BinaryPredicate):
                raise RuntimeError("MongoConnector::query: Not yet implemented")
            else:
                return where(ret, query.filters)
        else:
            raise RuntimeError("MongoConnector::query: Unable to query")
        return self.answer(query, ret)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

try:
    from ldap3                  import Server, Connection, SUBTREE, ALL_ATTRIBUTES, ALL
    from ldap3.core.exceptions  import LDAPInvalidFilterError
except ImportError:
    raise ImportError("LdapConnector requires python3-ldap3: please run: apt-get install python3-ldap3")

import operator
from .connector             import Connector
from .query                 import Query, ACTION_READ
from .binary_predicate      import BinaryPredicate
from .log                   import Log

class LdapConnector(Connector):
    def __init__(self, ldap_host, ldap_user = None, ldap_password = None, ldap_use_ssl = None):
        super().__init__()
        self.m_server     = Server(ldap_host, use_ssl = ldap_use_ssl, get_info = ALL)
        self.m_connection = Connection(self.m_server, ldap_user, ldap_password)
        self.m_connection.bind()

    def attributes(self, object :str) -> set:
        return {str(key) for key in self.m_connection.server.schema.attribute_types.keys()}

    def __enter__(self):
        self.m_connection.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        self.m_connection.unbind()

    @staticmethod
    def operator_to_ldap(op :str) -> str:
        if   op == operator.__eq__:  return "="
        elif op == operator.__and__: return "&"
        elif op == operator.__or__:  return "|"
        #elif op == "<=" or op == ">=" or op == "<" or op == ">" or op == "~": return op
        raise RuntimeError("LdapConnector::operator_to_ldap: op = %s not supported" % op)

    @staticmethod
    def binary_predicate_to_ldap(p :BinaryPredicate) -> str:
        if p.operator in ["&&", "||"]:
            # See http://ldap3.readthedocs.io/searches.html
            format = "(%(operator)s%(left)s%(right)s)"
        else:
            format = "%(left)s%(operator)s%(right)s"
        return format % {
            "left"     : LdapConnector.operand_to_ldap(p.left),
            "operator" : LdapConnector.operator_to_ldap(p.operator),
            "right"    : LdapConnector.operand_to_ldap(p.right)
        }

    @staticmethod
    def operand_to_ldap(operand) -> str:
        if isinstance(operand, BinaryPredicate):
            return "(%s)" % LdapConnector.binary_predicate_to_ldap(operand)
        elif isinstance(operand, (int, float, str)):
            # Do not add '...' or "..." for str
            return operand
        else:
            raise RuntimeError("Operand type not supported %s" % type(operand))

    @staticmethod
    def literal_from_ldap(b :bytes):
        try:
            return int(b)
        except ValueError:
            pass
        try:
            return float(b)
        except ValueError:
            pass
        try:
            return b.decode("utf-8")
        except ValueError:
            pass
        return b

    @staticmethod
    def sanitize_dict(d :dict) -> dict:
        sane = False
        for k,v in d.items():
            # raw values are always list
            if len(v) == 1:
                d[k] = LdapConnector.literal_from_ldap(v[0])
            else:
                d[k] = [LdapConnector.literal_from_ldap(x) for x in v]
            if d[k] == []:
                d[k] = None
            else:
                sane = True
        return d if sane == True else dict()

    def query(self, q :Query):
        super().query(q)
        entries = list()

        if q.action == ACTION_READ:
            if len(q.attributes) == 0:
                attributes = ALL_ATTRIBUTES
            else:
                attributes = q.attributes

            if q.filters == None:
                filter = "(objectClass=*)"
            else:
                filter = LdapConnector.operand_to_ldap(q.filters)

            try:
                if attributes != ALL_ATTRIBUTES:
                    attributes = set(attributes) & self.attributes(q.object)
                Log.info("--> LDAP: dn = %s filter = %s attributes = %s" % (q.object, filter, attributes))
                self.m_connection.search(
                    q.object,
                    filter,
                    search_scope = SUBTREE,
                    attributes = attributes
                )
            except LDAPInvalidFilterError as e:
                Log.error("LdapConnector::query: Invalid filter: %s" % filter)
                raise e

            for entry in self.m_connection.response:
                entry = LdapConnector.sanitize_dict(entry["raw_attributes"])
                if len(entry) > 0:
                    # Fix missing keys
                    if len(q.attributes) > 0:
                        for missing_key in (set(q.attributes) - set(entry.keys())):
                            entry[missing_key] = None
                    entries.append(dict(entry))
        else:
            raise RuntimeError("Not implemented")
        return self.answer(q, entries)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

def to_hashable(x):
    return tuple(to_hashable(elt) for elt in x) if isinstance(x, (set, list)) \
      else tuple((to_hashable(k), to_hashable(v)) for k, v in x.items()) if isinstance(x, dict) \
      else x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

def reverse_dict(d :dict) -> dict:
    return {v : k for (k, v) in d.items()}

def freeze_dict(d :dict):
    return {k : frozenset(v) if isinstance(v, set) else v for (k, v) in d.items()}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import sys
from pprint import pformat

DEBUG   = 0
INFO    = 1
WARNING = 2
ERROR   = 3

# Shell colors
DEFAULT = 0
RED     = 1
GREEN   = 2
YELLOW  = 3
BLUE    = 4
PINK    = 5
CYAN    = 6
GRAY    = 7

# Shell style
DEFAULT     = 0
BOLD        = 1
UNDERLINED  = 4
BLINKING    = 5
HIGHLIGHTED = 7

class Log:
    enable_print = False

    # TODO: The following static paramaters should be load from ~/.minifoldrc
    # TODO: dark / light colors
    with_color   = True
    log_level    = 0

    message_header = {
        DEBUG   : "DEBUG",
        INFO    : "INFO",
        WARNING : "WARNING",
        ERROR   : "ERROR",
    }

    message_color = {
        DEBUG   : CYAN,
        INFO    : GREEN,
        WARNING : YELLOW,
        ERROR   : RED,
    }

    @staticmethod
    def start_style(
        fg_color :int = None,
        bg_color :int = None,
        styles :list = list()
    ) -> str:
        styling = list()
        if fg_color != None: styling.append("3%d" % fg_color)
        if bg_color != None: styling.append("4%d" % bg_color)
        if styles: styling += styles
        return "\033[%sm" % ";".join(styling) if styling else ""

    @staticmethod
    def default_style() -> str:
        return "\033[0m"

    @classmethod
    def print(cls, message_type :int, message :str, file = sys.stderr):
        if cls.enable_print and message_type >= cls.log_level:
            color  = cls.message_color[message_type]
            header = cls.message_header[message_type]
            print(
                "%(start_style)s%(message)s%(end_style)s" % {
                    "start_style" : cls.start_style(fg_color = color),
                    "message"     : " ".join([header, message if isinstance(message, str) else pformat(message)]),
                    "end_style"   : cls.default_style()
                },
                file = file
            )

    @classmethod
    def debug(cls, s): cls.print(DEBUG, s)

    @classmethod
    def info(cls, s): cls.print(INFO, s)

    @classmethod
    def warning(cls, s): cls.print(WARNING, s)

    @classmethod
    def error(cls, s): cls.print(ERROR, s)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from itertools          import chain
from minifold.connector import Connector
from minifold.query     import Query, ACTION_READ

def union_gen(list_entries :list):
    """
    Iterator over a list of list of dicts.
    Args:
        list_entries: A list of lists of dicts.
    Returns:
        The resulting generator.
    """
    return chain(*list_entries)

def union(list_entries :list):
    """
    Make the union of several lists of entries.
    Use minifold.distinct() to remove duplicates.
    Args:
        list_entries: A list of lists of dicts.
    Returns:
        The resulting list of dicts.
    """
    return [entry for entry in union_gen(list_entries)]

class UnionConnector(Connector):
    def __init__(self, childs :list):
        self.m_childs = childs

    def attributes(self, object :str) -> set:
        attrs = set()
        for child in self.m_childs:
            try:
                attrs |= child.attributes(object)
            except:
                pass
        return attrs

    def query(self, query :Query) -> list:
        super().query(query)
        if query.action != ACTION_READ:
            raise ValueError("Invalid action in query %s" % query)
        return union([child.query(query) for child in self.m_childs])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from copy import deepcopy
from typing import List

ACTION_CREATE = 0
ACTION_READ   = 1
ACTION_UPDATE = 2
ACTION_DELETE = 3

SORT_ASC  = True
SORT_DESC = False

def action_to_str(action) -> str:
    if   action == ACTION_CREATE: return "INSERT"
    elif action == ACTION_READ  : return "SELECT"
    elif action == ACTION_UPDATE: return "UPDATE"
    elif action == ACTION_DELETE: return "DELETE"
    else: raise RuntimeError("action_to_str: invalid action %s" % action)

class Query:
    def __init__(self, action = ACTION_READ, object = "", attributes :List = list(), filters :List = None, offset = None, limit = None, sort_by = dict()):
        self.m_action = action
        self.m_object = object
        self.m_attributes = list(attributes)
        self.m_filters = filters

        # NOTE/TODO: the following specifiers are not supported by all the connectors
        self.m_offset = offset
        self.m_limit = limit
        self.m_sort_by = sort_by

    def copy(self):
        return deepcopy(self)

    @property
    def action(self) -> int:
        return self.m_action

    @action.setter
    def action(self, value :int):
        self.m_action = value

    @property
    def attributes(self) -> list:
        return self.m_attributes

    @attributes.setter
    def attributes(self, value :list):
        self.m_attributes = value

    @property
    def filters(self):
        return self.m_filters

    @filters.setter
    def filters(self, value):
        self.m_filters = value

    @property
    def limit(self) -> int:
        return self.m_limit

    @limit.setter
    def limit(self, value :int):
        self.m_limit = value

    @property
    def offset(self) -> int:
        return self.m_offset

    @offset.setter
    def offset(self, value :int):
        self.m_offset = value

    @property
    def sort_by(self) -> dict:
        return self.m_sort_by

    @sort_by.setter
    def sort_by(self, value :dict):
        self.m_sort_by = value

    @property
    def object(self) -> str:
        return self.m_object

    def __str__(self) -> str:
        return "%(action)s%(attributes)s%(object)s%(filters)s%(limit)s%(offset)s%(sort_by)s" % {
            "action"     : action_to_str(self.action),
            "attributes" : " %s" % ", ".join(sorted(set(self.attributes))) if len(self.attributes) > 0 else " *",
            "object"     : " FROM %s" % self.object   if self.object  else "",
            "filters"    : " WHERE %s" % self.filters if self.filters else "",
            "limit"      : " LIMIT %s" % self.limit   if self.limit   else "",
            "offset"     : " OFFSET %s" % self.offset if self.offset  else "",
            "sort_by"    : " SORT BY %s" % ", ".join([
                    "%s %s" % (
                        attribute,
                        "ASC" if sort_asc == SORT_ASC else "DESC"
                    ) for attribute, sort_asc in self.sort_by.items()
                ]) if self.sort_by else "",
        }




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import re
from .where     import where

def equals(x, y) -> bool:
    return x == y
def contains(x, y) -> bool:
    return x in y
def lower_case_equals(x :str, y :str) -> bool:
    return x.lower() == y.lower()
def lower_case_contains(x :str, y :str) -> bool:
    return x.lower() in y.lower()

def contains_words(word, sentence, ignore_case = True):
    pattern = ".*\\b(%s)\\b.*" % word
    flags = re.IGNORECASE if ignore_case else 0
    regex = re.compile(pattern, flags)
    return bool(regex.match(sentence))

class SearchFilter:
    def __init__(self, search_values :list, attributes :list, match):
        self.search_values = search_values
        self.attributes = attributes
        self.match = match

    def __call__(self, entry :dict) -> bool:
        for search_value in self.search_values:
            for attribute in self.attributes:
                if self.match(search_value, entry[attribute]):
                    return True
        return False

def search(
    entries       :list,
    attributes    :list,
    search_values :list,
    match = equals
) -> list:
    return where(entries, SearchFilter(search_values, attributes, match))



#! /usr/bin/env python3
# This file is copied from the scholar.py project.
# https://github.com/ckreibich/scholar.py
# (BSD)

"""
This module provides classes for querying Google Scholar and parsing
returned results. It currently *only* processes the first results
page. It is not a recursive crawler.
"""
import optparse
import os
import re
import sys
import warnings

try:
    # Try importing for Python 3
    # pylint: disable-msg=F0401
    # pylint: disable-msg=E0611
    from urllib.request import HTTPCookieProcessor, Request, build_opener
    from urllib.parse import quote, unquote
    from http.cookiejar import MozillaCookieJar
except ImportError:
    # Fallback for Python 2
    from urllib2 import Request, build_opener, HTTPCookieProcessor
    from urllib import quote, unquote
    from cookielib import MozillaCookieJar

# Import BeautifulSoup -- try 4 first, fall back to older
try:
    from bs4 import BeautifulSoup, NavigableString
except ImportError:
    try:
        from BeautifulSoup import BeautifulSoup, NavigableString
    except ImportError:
        print('We need BeautifulSoup, sorry...')
        sys.exit(1)

# Support unicode in both Python 2 and 3. In Python 3, unicode is str.
if sys.version_info[0] == 3:
    unicode = str # pylint: disable-msg=W0622
    encode = lambda s: unicode(s) # pylint: disable-msg=C0103
else:
    def encode(s):
        if isinstance(s, basestring):
            return s.encode('utf-8') # pylint: disable-msg=C0103
        else:
            return str(s)


class Error(Exception):
    """Base class for any Scholar error."""


class FormatError(Error):
    """A query argument or setting was formatted incorrectly."""


class QueryArgumentError(Error):
    """A query did not have a suitable set of arguments."""


class SoupKitchen(object):
    """Factory for creating BeautifulSoup instances."""

    @staticmethod
    def make_soup(markup, parser=None):
        """Factory method returning a BeautifulSoup instance. The created
        instance will use a parser of the given name, if supported by
        the underlying BeautifulSoup instance.
        """
        if 'bs4' in sys.modules:
            # We support parser specification. If the caller didn't
            # specify one, leave it to BeautifulSoup to pick the most
            # suitable one, but suppress the user warning that asks to
            # select the most suitable parser ... which BS then
            # selects anyway.
            if parser is None:
                warnings.filterwarnings('ignore', 'No parser was explicitly specified')
            return BeautifulSoup(markup, parser)

        return BeautifulSoup(markup)

class ScholarConf(object):
    """Helper class for global settings."""

    VERSION = '2.10'
    LOG_LEVEL = 1
    MAX_PAGE_RESULTS = 10 # Current default for per-page results
    SCHOLAR_SITE = 'http://scholar.google.com'

    # USER_AGENT = 'Mozilla/5.0 (X11; U; FreeBSD i386; en-US; rv:1.9.2.9) Gecko/20100913 Firefox/3.6.9'
    # Let's update at this point (3/14):
    USER_AGENT = 'Mozilla/5.0 (X11; Linux x86_64; rv:27.0) Gecko/20100101 Firefox/27.0'

    # If set, we will use this file to read/save cookies to enable
    # cookie use across sessions.
    COOKIE_JAR_FILE = None

class ScholarUtils(object):
    """A wrapper for various utensils that come in handy."""

    LOG_LEVELS = {'error': 1,
                  'warn':  2,
                  'info':  3,
                  'debug': 4}

    @staticmethod
    def ensure_int(arg, msg=None):
        try:
            return int(arg)
        except ValueError:
            raise FormatError(msg)

    @staticmethod
    def log(level, msg):
        if level not in ScholarUtils.LOG_LEVELS.keys():
            return
        if ScholarUtils.LOG_LEVELS[level] > ScholarConf.LOG_LEVEL:
            return
        sys.stderr.write('[%5s]  %s' % (level.upper(), msg + '\n'))
        sys.stderr.flush()


class ScholarArticle(object):
    """
    A class representing articles listed on Google Scholar.  The class
    provides basic dictionary-like behavior.
    """
    def __init__(self):
        # The triplets for each keyword correspond to (1) the actual
        # value, (2) a user-suitable label for the item, and (3) an
        # ordering index:
        self.attrs = {
            'title':         [None, 'Title',          0],
            'url':           [None, 'URL',            1],
            'year':          [None, 'Year',           2],
            'num_citations': [0,    'Citations',      3],
            'num_versions':  [0,    'Versions',       4],
            'cluster_id':    [None, 'Cluster ID',     5],
            'url_pdf':       [None, 'PDF link',       6],
            'url_citations': [None, 'Citations list', 7],
            'url_versions':  [None, 'Versions list',  8],
            'url_citation':  [None, 'Citation link',  9],
            'excerpt':       [None, 'Excerpt',       10],
        }

        # The citation data in one of the standard export formats,
        # e.g. BibTeX.
        self.citation_data = None

    def __getitem__(self, key):
        if key in self.attrs:
            return self.attrs[key][0]
        return None

    def __len__(self):
        return len(self.attrs)

    def __setitem__(self, key, item):
        if key in self.attrs:
            self.attrs[key][0] = item
        else:
            self.attrs[key] = [item, key, len(self.attrs)]

    def __delitem__(self, key):
        if key in self.attrs:
            del self.attrs[key]

    def set_citation_data(self, citation_data):
        self.citation_data = citation_data

    def as_txt(self):
        # Get items sorted in specified order:
        items = sorted(list(self.attrs.values()), key=lambda item: item[2])
        # Find largest label length:
        max_label_len = max([len(str(item[1])) for item in items])
        fmt = '%%%ds %%s' % max_label_len
        res = []
        for item in items:
            if item[0] is not None:
                res.append(fmt % (item[1], item[0]))
        return '\n'.join(res)

    def as_csv(self, header=False, sep='|'):
        # Get keys sorted in specified order:
        keys = [pair[0] for pair in \
                sorted([(key, val[2]) for key, val in list(self.attrs.items())],
                       key=lambda pair: pair[1])]
        res = []
        if header:
            res.append(sep.join(keys))
        res.append(sep.join([unicode(self.attrs[key][0]) for key in keys]))
        return '\n'.join(res)

    def as_citation(self):
        """
        Reports the article in a standard citation format. This works only
        if you have configured the querier to retrieve a particular
        citation export format. (See ScholarSettings.)
        """
        return self.citation_data or ''


class ScholarArticleParser(object):
    """
    ScholarArticleParser can parse HTML document strings obtained from
    Google Scholar. This is a base class; concrete implementations
    adapting to tweaks made by Google over time follow below.
    """
    def __init__(self, site=None):
        self.soup = None
        self.article = None
        self.site = site or ScholarConf.SCHOLAR_SITE
        self.year_re = re.compile(r'\b(?:20|19)\d{2}\b')

    def handle_article(self, art):
        """
        The parser invokes this callback on each article parsed
        successfully.  In this base class, the callback does nothing.
        """

    def handle_num_results(self, num_results):
        """
        The parser invokes this callback if it determines the overall
        number of results, as reported on the parsed results page. The
        base class implementation does nothing.
        """

    def parse(self, html):
        """
        This method initiates parsing of HTML content, cleans resulting
        content as needed, and notifies the parser instance of
        resulting instances via the handle_article callback.
        """
        self.soup = SoupKitchen.make_soup(html)

        # This parses any global, non-itemized attributes from the page.
        self._parse_globals()

        # Now parse out listed articles:
        for div in self.soup.findAll(ScholarArticleParser._tag_results_checker):
            self._parse_article(div)
            self._clean_article()
            if self.article['title']:
                self.handle_article(self.article)

    def _clean_article(self):
        """
        This gets invoked after we have parsed an article, to do any
        needed cleanup/polishing before we hand off the resulting
        article.
        """
        if self.article['title']:
            self.article['title'] = self.article['title'].strip()

    def _parse_globals(self):
        tag = self.soup.find(name='div', attrs={'id': 'gs_ab_md'})
        if tag is not None:
            raw_text = tag.findAll(text=True)
            # raw text is a list because the body contains <b> etc
            if raw_text is not None and len(raw_text) > 0:
                try:
                    num_results = raw_text[0].split()[1]
                    # num_results may now contain commas to separate
                    # thousands, strip:
                    num_results = num_results.replace(',', '')
                    num_results = int(num_results)
                    self.handle_num_results(num_results)
                except (IndexError, ValueError):
                    pass

    def _parse_article(self, div):
        self.article = ScholarArticle()

        for tag in div:
            if not hasattr(tag, 'name'):
                continue

            if tag.name == 'div' and self._tag_has_class(tag, 'gs_rt') and \
                    tag.h3 and tag.h3.a:
                self.article['title'] = ''.join(tag.h3.a.findAll(text=True))
                self.article['url'] = self._path2url(tag.h3.a['href'])
                if self.article['url'].endswith('.pdf'):
                    self.article['url_pdf'] = self.article['url']

            if tag.name == 'font':
                for tag2 in tag:
                    if not hasattr(tag2, 'name'):
                        continue
                    if tag2.name == 'span' and \
                       self._tag_has_class(tag2, 'gs_fl'):
                        self._parse_links(tag2)

    def _parse_links(self, span):
        for tag in span:
            if not hasattr(tag, 'name'):
                continue
            if tag.name != 'a' or tag.get('href') is None:
                continue

            if tag.get('href').startswith('/scholar?cites'):
                if hasattr(tag, 'string') and tag.string.startswith('Cited by'):
                    self.article['num_citations'] = \
                        self._as_int(tag.string.split()[-1])

                # Weird Google Scholar behavior here: if the original
                # search query came with a number-of-results limit,
                # then this limit gets propagated to the URLs embedded
                # in the results page as well. Same applies to
                # versions URL in next if-block.
                self.article['url_citations'] = \
                    self._strip_url_arg('num', self._path2url(tag.get('href')))

                # We can also extract the cluster ID from the versions
                # URL. Note that we know that the string contains "?",
                # from the above if-statement.
                args = self.article['url_citations'].split('?', 1)[1]
                for arg in args.split('&'):
                    if arg.startswith('cites='):
                        self.article['cluster_id'] = arg[6:]

            if tag.get('href').startswith('/scholar?cluster'):
                if hasattr(tag, 'string') and tag.string.startswith('All '):
                    self.article['num_versions'] = \
                        self._as_int(tag.string.split()[1])
                self.article['url_versions'] = \
                    self._strip_url_arg('num', self._path2url(tag.get('href')))

            if tag.getText().startswith('Import'):
                self.article['url_citation'] = self._path2url(tag.get('href'))


    @staticmethod
    def _tag_has_class(tag, klass):
        """
        This predicate function checks whether a BeatifulSoup Tag instance
        has a class attribute.
        """
        res = tag.get('class') or []
        if type(res) != list:
            # BeautifulSoup 3 can return e.g. 'gs_md_wp gs_ttss',
            # so split -- conveniently produces a list in any case
            res = res.split()
        return klass in res

    @staticmethod
    def _tag_results_checker(tag):
        return tag.name == 'div' \
            and ScholarArticleParser._tag_has_class(tag, 'gs_r')

    @staticmethod
    def _as_int(obj):
        try:
            return int(obj)
        except ValueError:
            return None

    def _path2url(self, path):
        """Helper, returns full URL in case path isn't one."""
        if path.startswith('http://'):
            return path
        if not path.startswith('/'):
            path = '/' + path
        return self.site + path

    def _strip_url_arg(self, arg, url):
        """Helper, removes a URL-encoded argument, if present."""
        parts = url.split('?', 1)
        if len(parts) != 2:
            return url
        res = []
        for part in parts[1].split('&'):
            if not part.startswith(arg + '='):
                res.append(part)
        return parts[0] + '?' + '&'.join(res)

class ScholarArticleParser120726(ScholarArticleParser):
    """
    This class reflects update to the Scholar results page layout that
    Google made 07/26/12.
    """
    def _parse_article(self, div):
        self.article = ScholarArticle()

        for tag in div:
            if not hasattr(tag, 'name'):
                continue
            if str(tag).lower().find('.pdf'):
#                if tag.find('div', class_='gs_ttss'):
                if not isinstance(tag, NavigableString) and tag.find('div', class_='gs_ttss'):
                    self._parse_links(tag.find('div', {'class': 'gs_ttss'}))

            if tag.name == 'div' and self._tag_has_class(tag, 'gs_ri'):
                # There are (at least) two formats here. In the first
                # one, we have a link, e.g.:
                #
                # <h3 class="gs_rt">
                #   <a href="http://dl.acm.org/citation.cfm?id=972384" class="yC0">
                #     <b>Honeycomb</b>: creating intrusion detection signatures using
                #        honeypots
                #   </a>
                # </h3>
                #
                # In the other, there's no actual link -- it's what
                # Scholar renders as "CITATION" in the HTML:
                #
                # <h3 class="gs_rt">
                #   <span class="gs_ctu">
                #     <span class="gs_ct1">[CITATION]</span>
                #     <span class="gs_ct2">[C]</span>
                #   </span>
                #   <b>Honeycomb</b> automated ids signature creation using honeypots
                # </h3>
                #
                # We now distinguish the two.
                try:
                    atag = tag.h3.a
                    self.article['title'] = ''.join(atag.findAll(text=True))
                    self.article['url'] = self._path2url(atag['href'])
                    if self.article['url'].endswith('.pdf'):
                        self.article['url_pdf'] = self.article['url']
                except:
                    # Remove a few spans that have unneeded content (e.g. [CITATION])
                    for span in tag.h3.findAll(name='span'):
                        span.clear()
                    self.article['title'] = ''.join(tag.h3.findAll(text=True))

                if tag.find('div', {'class': 'gs_a'}):
                    year = self.year_re.findall(tag.find('div', {'class': 'gs_a'}).text)
                    self.article['year'] = year[0] if len(year) > 0 else None

                if tag.find('div', {'class': 'gs_fl'}):
                    self._parse_links(tag.find('div', {'class': 'gs_fl'}))

                if tag.find('div', {'class': 'gs_rs'}):
                    # These are the content excerpts rendered into the results.
                    raw_text = tag.find('div', {'class': 'gs_rs'}).findAll(text=True)
                    if len(raw_text) > 0:
                        raw_text = ''.join(raw_text)
                        raw_text = raw_text.replace('\n', '')
                        self.article['excerpt'] = raw_text


class ScholarQuery(object):
    """
    The base class for any kind of results query we send to Scholar.
    """
    def __init__(self):
        self.url = None

        # The number of results requested from Scholar -- not the
        # total number of results it reports (the latter gets stored
        # in attrs, see below).
        self.num_results = None

        # Queries may have global result attributes, similar to
        # per-article attributes in ScholarArticle. The exact set of
        # attributes may differ by query type, but they all share the
        # basic data structure:
        self.attrs = {}

    def set_num_page_results(self, num_page_results):
        self.num_results = ScholarUtils.ensure_int(
            num_page_results,
            'maximum number of results on page must be numeric')

    def get_url(self):
        """
        Returns a complete, submittable URL string for this particular
        query instance. The URL and its arguments will vary depending
        on the query.
        """
        return None

    def _add_attribute_type(self, key, label, default_value=None):
        """
        Adds a new type of attribute to the list of attributes
        understood by this query. Meant to be used by the constructors
        in derived classes.
        """
        if len(self.attrs) == 0:
            self.attrs[key] = [default_value, label, 0]
            return
        idx = max([item[2] for item in self.attrs.values()]) + 1
        self.attrs[key] = [default_value, label, idx]

    def __getitem__(self, key):
        """Getter for attribute value. Returns None if no such key."""
        if key in self.attrs:
            return self.attrs[key][0]
        return None

    def __setitem__(self, key, item):
        """Setter for attribute value. Does nothing if no such key."""
        if key in self.attrs:
            self.attrs[key][0] = item

    def _parenthesize_phrases(self, query):
        """
        Turns a query string containing comma-separated phrases into a
        space-separated list of tokens, quoted if containing
        whitespace. For example, input

          'some words, foo, bar'

        becomes

          '"some words" foo bar'

        This comes in handy during the composition of certain queries.
        """
        if query.find(',') < 0:
            return query
        phrases = []
        for phrase in query.split(','):
            phrase = phrase.strip()
            if phrase.find(' ') > 0:
                phrase = '"' + phrase + '"'
            phrases.append(phrase)
        return ' '.join(phrases)


class ClusterScholarQuery(ScholarQuery):
    """
    This version just pulls up an article cluster whose ID we already
    know about.
    """
    SCHOLAR_CLUSTER_URL = ScholarConf.SCHOLAR_SITE + '/scholar?' \
        + 'cluster=%(cluster)s' \
        + '%(num)s'

    def __init__(self, cluster=None):
        ScholarQuery.__init__(self)
        self._add_attribute_type('num_results', 'Results', 0)
        self.cluster = None
        self.set_cluster(cluster)

    def set_cluster(self, cluster):
        """
        Sets search to a Google Scholar results cluster ID.
        """
        msg = 'cluster ID must be numeric'
        self.cluster = ScholarUtils.ensure_int(cluster, msg)

    def get_url(self):
        if self.cluster is None:
            raise QueryArgumentError('cluster query needs cluster ID')

        urlargs = {'cluster': self.cluster }

        for key, val in urlargs.items():
            urlargs[key] = quote(encode(val))

        # The following URL arguments must not be quoted, or the
        # server will not recognize them:
        urlargs['num'] = ('&num=%d' % self.num_results
                          if self.num_results is not None else '')

        return self.SCHOLAR_CLUSTER_URL % urlargs


class SearchScholarQuery(ScholarQuery):
    """
    This version represents the search query parameters the user can
    configure on the Scholar website, in the advanced search options.
    """
    SCHOLAR_QUERY_URL = ScholarConf.SCHOLAR_SITE + '/scholar?' \
        + 'as_q=%(words)s' \
        + '&as_epq=%(phrase)s' \
        + '&as_oq=%(words_some)s' \
        + '&as_eq=%(words_none)s' \
        + '&as_occt=%(scope)s' \
        + '&as_sauthors=%(authors)s' \
        + '&as_publication=%(pub)s' \
        + '&as_ylo=%(ylo)s' \
        + '&as_yhi=%(yhi)s' \
        + '&as_vis=%(citations)s' \
        + '&btnG=&hl=en' \
        + '%(num)s' \
        + '&as_sdt=%(patents)s%%2C5'

    def __init__(self):
        ScholarQuery.__init__(self)
        self._add_attribute_type('num_results', 'Results', 0)
        self.words = None # The default search behavior
        self.words_some = None # At least one of those words
        self.words_none = None # None of these words
        self.phrase = None
        self.scope_title = False # If True, search in title only
        self.author = None
        self.pub = None
        self.timeframe = [None, None]
        self.include_patents = True
        self.include_citations = True

    def set_words(self, words):
        """Sets words that *all* must be found in the result."""
        self.words = words

    def set_words_some(self, words):
        """Sets words of which *at least one* must be found in result."""
        self.words_some = words

    def set_words_none(self, words):
        """Sets words of which *none* must be found in the result."""
        self.words_none = words

    def set_phrase(self, phrase):
        """Sets phrase that must be found in the result exactly."""
        self.phrase = phrase

    def set_scope(self, title_only):
        """
        Sets Boolean indicating whether to search entire article or title
        only.
        """
        self.scope_title = title_only

    def set_author(self, author):
        """Sets names that must be on the result's author list."""
        self.author = author

    def set_pub(self, pub):
        """Sets the publication in which the result must be found."""
        self.pub = pub

    def set_timeframe(self, start=None, end=None):
        """
        Sets timeframe (in years as integer) in which result must have
        appeared. It's fine to specify just start or end, or both.
        """
        if start:
            start = ScholarUtils.ensure_int(start)
        if end:
            end = ScholarUtils.ensure_int(end)
        self.timeframe = [start, end]

    def set_include_citations(self, yesorno):
        self.include_citations = yesorno

    def set_include_patents(self, yesorno):
        self.include_patents = yesorno

    def get_url(self):
        if self.words is None and self.words_some is None \
           and self.words_none is None and self.phrase is None \
           and self.author is None and self.pub is None \
           and self.timeframe[0] is None and self.timeframe[1] is None:
            raise QueryArgumentError('search query needs more parameters')

        # If we have some-words or none-words lists, we need to
        # process them so GS understands them. For simple
        # space-separeted word lists, there's nothing to do. For lists
        # of phrases we have to ensure quotations around the phrases,
        # separating them by whitespace.
        words_some = None
        words_none = None

        if self.words_some:
            words_some = self._parenthesize_phrases(self.words_some)
        if self.words_none:
            words_none = self._parenthesize_phrases(self.words_none)

        urlargs = {'words': self.words or '',
                   'words_some': words_some or '',
                   'words_none': words_none or '',
                   'phrase': self.phrase or '',
                   'scope': 'title' if self.scope_title else 'any',
                   'authors': self.author or '',
                   'pub': self.pub or '',
                   'ylo': self.timeframe[0] or '',
                   'yhi': self.timeframe[1] or '',
                   'patents': '0' if self.include_patents else '1',
                   'citations': '0' if self.include_citations else '1'}

        for key, val in urlargs.items():
            urlargs[key] = quote(encode(val))

        # The following URL arguments must not be quoted, or the
        # server will not recognize them:
        urlargs['num'] = ('&num=%d' % self.num_results
                          if self.num_results is not None else '')

        return self.SCHOLAR_QUERY_URL % urlargs


class ScholarSettings(object):
    """
    This class lets you adjust the Scholar settings for your
    session. It's intended to mirror the features tunable in the
    Scholar Settings pane, but right now it's a bit basic.
    """
    CITFORM_NONE = 0
    CITFORM_REFWORKS = 1
    CITFORM_REFMAN = 2
    CITFORM_ENDNOTE = 3
    CITFORM_BIBTEX = 4

    def __init__(self):
        self.citform = 0 # Citation format, default none
        self.per_page_results = None
        self._is_configured = False

    def set_citation_format(self, citform):
        citform = ScholarUtils.ensure_int(citform)
        if citform < 0 or citform > self.CITFORM_BIBTEX:
            raise FormatError('citation format invalid, is "%s"'
                              % citform)
        self.citform = citform
        self._is_configured = True

    def set_per_page_results(self, per_page_results):
        self.per_page_results = ScholarUtils.ensure_int(
            per_page_results, 'page results must be integer')
        self.per_page_results = min(
            self.per_page_results, ScholarConf.MAX_PAGE_RESULTS)
        self._is_configured = True

    def is_configured(self):
        return self._is_configured


class ScholarQuerier(object):
    """
    ScholarQuerier instances can conduct a search on Google Scholar
    with subsequent parsing of the resulting HTML content.  The
    articles found are collected in the articles member, a list of
    ScholarArticle instances.
    """

    # Default URLs for visiting and submitting Settings pane, as of 3/14
    GET_SETTINGS_URL = ScholarConf.SCHOLAR_SITE + '/scholar_settings?' \
        + 'sciifh=1&hl=en&as_sdt=0,5'

    SET_SETTINGS_URL = ScholarConf.SCHOLAR_SITE + '/scholar_setprefs?' \
        + 'q=' \
        + '&scisig=%(scisig)s' \
        + '&inststart=0' \
        + '&as_sdt=1,5' \
        + '&as_sdtp=' \
        + '&num=%(num)s' \
        + '&scis=%(scis)s' \
        + '%(scisf)s' \
        + '&hl=en&lang=all&instq=&inst=569367360547434339&save='

    # Older URLs:
    # ScholarConf.SCHOLAR_SITE + '/scholar?q=%s&hl=en&btnG=Search&as_sdt=2001&as_sdtp=on

    class Parser(ScholarArticleParser120726):
        def __init__(self, querier):
            ScholarArticleParser120726.__init__(self)
            self.querier = querier

        def handle_num_results(self, num_results):
            if self.querier is not None and self.querier.query is not None:
                self.querier.query['num_results'] = num_results

        def handle_article(self, art):
            self.querier.add_article(art)

    def __init__(self):
        self.articles = []
        self.query = None
        self.cjar = MozillaCookieJar()

        # If we have a cookie file, load it:
        if ScholarConf.COOKIE_JAR_FILE and \
           os.path.exists(ScholarConf.COOKIE_JAR_FILE):
            try:
                self.cjar.load(ScholarConf.COOKIE_JAR_FILE,
                               ignore_discard=True)
                ScholarUtils.log('info', 'loaded cookies file')
            except Exception as msg:
                ScholarUtils.log('warn', 'could not load cookies file: %s' % msg)
                self.cjar = MozillaCookieJar() # Just to be safe

        self.opener = build_opener(HTTPCookieProcessor(self.cjar))
        self.settings = None # Last settings object, if any

    def apply_settings(self, settings):
        """
        Applies settings as provided by a ScholarSettings instance.
        """
        if settings is None or not settings.is_configured():
            return True

        self.settings = settings

        # This is a bit of work. We need to actually retrieve the
        # contents of the Settings pane HTML in order to extract
        # hidden fields before we can compose the query for updating
        # the settings.
        html = self._get_http_response(url=self.GET_SETTINGS_URL,
                                       log_msg='dump of settings form HTML',
                                       err_msg='requesting settings failed')
        if html is None:
            return False

        # Now parse the required stuff out of the form. We require the
        # "scisig" token to make the upload of our settings acceptable
        # to Google.
        soup = SoupKitchen.make_soup(html)

        tag = soup.find(name='form', attrs={'id': 'gs_settings_form'})
        if tag is None:
            ScholarUtils.log('info', 'parsing settings failed: no form')
            return False

        tag = tag.find('input', attrs={'type':'hidden', 'name':'scisig'})
        if tag is None:
            ScholarUtils.log('info', 'parsing settings failed: scisig')
            return False

        urlargs = {'scisig': tag['value'],
                   'num': settings.per_page_results,
                   'scis': 'no',
                   'scisf': ''}

        if settings.citform != 0:
            urlargs['scis'] = 'yes'
            urlargs['scisf'] = '&scisf=%d' % settings.citform

        html = self._get_http_response(url=self.SET_SETTINGS_URL % urlargs,
                                       log_msg='dump of settings result HTML',
                                       err_msg='applying setttings failed')
        if html is None:
            return False

        ScholarUtils.log('info', 'settings applied')
        return True

    def send_query(self, query):
        """
        This method initiates a search query (a ScholarQuery instance)
        with subsequent parsing of the response.
        """
        self.clear_articles()
        self.query = query

        html = self._get_http_response(url=query.get_url(),
                                       log_msg='dump of query response HTML',
                                       err_msg='results retrieval failed')
        if html is None:
            return

        self.parse(html)

    def get_citation_data(self, article):
        """
        Given an article, retrieves citation link. Note, this requires that
        you adjusted the settings to tell Google Scholar to actually
        provide this information, *prior* to retrieving the article.
        """
        if article['url_citation'] is None:
            return False
        if article.citation_data is not None:
            return True

        ScholarUtils.log('info', 'retrieving citation export data')
        data = self._get_http_response(url=article['url_citation'],
                                       log_msg='citation data response',
                                       err_msg='requesting citation data failed')
        if data is None:
            return False

        article.set_citation_data(data)
        return True

    def parse(self, html):
        """
        This method allows parsing of provided HTML content.
        """
        parser = self.Parser(self)
        parser.parse(html)

    def add_article(self, art):
        self.get_citation_data(art)
        self.articles.append(art)

    def clear_articles(self):
        """Clears any existing articles stored from previous queries."""
        self.articles = []

    def save_cookies(self):
        """
        This stores the latest cookies we're using to disk, for reuse in a
        later session.
        """
        if ScholarConf.COOKIE_JAR_FILE is None:
            return False
        try:
            self.cjar.save(ScholarConf.COOKIE_JAR_FILE,
                           ignore_discard=True)
            ScholarUtils.log('info', 'saved cookies file')
            return True
        except Exception as msg:
            ScholarUtils.log('warn', 'could not save cookies file: %s' % msg)
            return False

    def _get_http_response(self, url, log_msg=None, err_msg=None):
        """
        Helper method, sends HTTP request and returns response payload.
        """
        if log_msg is None:
            log_msg = 'HTTP response data follow'
        if err_msg is None:
            err_msg = 'request failed'
        try:
            ScholarUtils.log('info', 'requesting %s' % unquote(url))

            req = Request(url=url, headers={'User-Agent': ScholarConf.USER_AGENT})
            hdl = self.opener.open(req)
            html = hdl.read()

            ScholarUtils.log('debug', log_msg)
            ScholarUtils.log('debug', '>>>>' + '-'*68)
            ScholarUtils.log('debug', 'url: %s' % hdl.geturl())
            ScholarUtils.log('debug', 'result: %s' % hdl.getcode())
            ScholarUtils.log('debug', 'headers:\n' + str(hdl.info()))
            ScholarUtils.log('debug', 'data:\n' + html.decode('utf-8')) # For Python 3
            ScholarUtils.log('debug', '<<<<' + '-'*68)

            return html
        except Exception as err:
            ScholarUtils.log('info', err_msg + ': %s' % err)
            return None


def txt(querier, with_globals):
    if with_globals:
        # If we have any articles, check their attribute labels to get
        # the maximum length -- makes for nicer alignment.
        max_label_len = 0
        if len(querier.articles) > 0:
            items = sorted(list(querier.articles[0].attrs.values()),
                           key=lambda item: item[2])
            max_label_len = max([len(str(item[1])) for item in items])

        # Get items sorted in specified order:
        items = sorted(list(querier.query.attrs.values()), key=lambda item: item[2])
        # Find largest label length:
        max_label_len = max([len(str(item[1])) for item in items] + [max_label_len])
        fmt = '[G] %%%ds %%s' % max(0, max_label_len-4)
        for item in items:
            if item[0] is not None:
                print(fmt % (item[1], item[0]))
        if len(items) > 0:
            print

    articles = querier.articles
    for art in articles:
        print(encode(art.as_txt()) + '\n')

def csv(querier, header=False, sep='|'):
    articles = querier.articles
    for art in articles:
        result = art.as_csv(header=header, sep=sep)
        print(encode(result))
        header = False

def citation_export(querier):
    articles = querier.articles
    for art in articles:
        print(art.as_citation() + '\n')


def main():
    usage = """scholar.py [options] <query string>
A command-line interface to Google Scholar.

Examples:

# Retrieve one article written by Einstein on quantum theory:
scholar.py -c 1 --author "albert einstein" --phrase "quantum theory"

# Retrieve a BibTeX entry for that quantum theory paper:
scholar.py -c 1 -C 17749203648027613321 --citation bt

# Retrieve five articles written by Einstein after 1970 where the title
# does not contain the words "quantum" and "theory":
scholar.py -c 5 -a "albert einstein" -t --none "quantum theory" --after 1970"""

    fmt = optparse.IndentedHelpFormatter(max_help_position=50, width=100)
    parser = optparse.OptionParser(usage=usage, formatter=fmt)
    group = optparse.OptionGroup(parser, 'Query arguments',
                                 'These options define search query arguments and parameters.')
    group.add_option('-a', '--author', metavar='AUTHORS', default=None,
                     help='Author name(s)')
    group.add_option('-A', '--all', metavar='WORDS', default=None, dest='allw',
                     help='Results must contain all of these words')
    group.add_option('-s', '--some', metavar='WORDS', default=None,
                     help='Results must contain at least one of these words. Pass arguments in form -s "foo bar baz" for simple words, and -s "a phrase, another phrase" for phrases')
    group.add_option('-n', '--none', metavar='WORDS', default=None,
                     help='Results must contain none of these words. See -s|--some re. formatting')
    group.add_option('-p', '--phrase', metavar='PHRASE', default=None,
                     help='Results must contain exact phrase')
    group.add_option('-t', '--title-only', action='store_true', default=False,
                     help='Search title only')
    group.add_option('-P', '--pub', metavar='PUBLICATIONS', default=None,
                     help='Results must have appeared in this publication')
    group.add_option('--after', metavar='YEAR', default=None,
                     help='Results must have appeared in or after given year')
    group.add_option('--before', metavar='YEAR', default=None,
                     help='Results must have appeared in or before given year')
    group.add_option('--no-patents', action='store_true', default=False,
                     help='Do not include patents in results')
    group.add_option('--no-citations', action='store_true', default=False,
                     help='Do not include citations in results')
    group.add_option('-C', '--cluster-id', metavar='CLUSTER_ID', default=None,
                     help='Do not search, just use articles in given cluster ID')
    group.add_option('-c', '--count', type='int', default=None,
                     help='Maximum number of results')
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, 'Output format',
                                 'These options control the appearance of the results.')
    group.add_option('--txt', action='store_true',
                     help='Print article data in text format (default)')
    group.add_option('--txt-globals', action='store_true',
                     help='Like --txt, but first print global results too')
    group.add_option('--csv', action='store_true',
                     help='Print article data in CSV form (separator is "|")')
    group.add_option('--csv-header', action='store_true',
                     help='Like --csv, but print header with column names')
    group.add_option('--citation', metavar='FORMAT', default=None,
                     help='Print article details in standard citation format. Argument Must be one of "bt" (BibTeX), "en" (EndNote), "rm" (RefMan), or "rw" (RefWorks).')
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, 'Miscellaneous')
    group.add_option('--cookie-file', metavar='FILE', default=None,
                     help='File to use for cookie storage. If given, will read any existing cookies if found at startup, and save resulting cookies in the end.')
    group.add_option('-d', '--debug', action='count', default=0,
                     help='Enable verbose logging to stderr. Repeated options increase detail of debug output.')
    group.add_option('-v', '--version', action='store_true', default=False,
                     help='Show version information')
    parser.add_option_group(group)

    options, _ = parser.parse_args()

    # Show help if we have neither keyword search nor author name
    if len(sys.argv) == 1:
        parser.print_help()
        return 1

    if options.debug > 0:
        options.debug = min(options.debug, ScholarUtils.LOG_LEVELS['debug'])
        ScholarConf.LOG_LEVEL = options.debug
        ScholarUtils.log('info', 'using log level %d' % ScholarConf.LOG_LEVEL)

    if options.version:
        print('This is scholar.py %s.' % ScholarConf.VERSION)
        return 0

    if options.cookie_file:
        ScholarConf.COOKIE_JAR_FILE = options.cookie_file

    # Sanity-check the options: if they include a cluster ID query, it
    # makes no sense to have search arguments:
    if options.cluster_id is not None:
        if options.author or options.allw or options.some or options.none \
           or options.phrase or options.title_only or options.pub \
           or options.after or options.before:
            print('Cluster ID queries do not allow additional search arguments.')
            return 1

    querier = ScholarQuerier()
    settings = ScholarSettings()

    if options.citation == 'bt':
        settings.set_citation_format(ScholarSettings.CITFORM_BIBTEX)
    elif options.citation == 'en':
        settings.set_citation_format(ScholarSettings.CITFORM_ENDNOTE)
    elif options.citation == 'rm':
        settings.set_citation_format(ScholarSettings.CITFORM_REFMAN)
    elif options.citation == 'rw':
        settings.set_citation_format(ScholarSettings.CITFORM_REFWORKS)
    elif options.citation is not None:
        print('Invalid citation link format, must be one of "bt", "en", "rm", or "rw".')
        return 1

    querier.apply_settings(settings)

    if options.cluster_id:
        query = ClusterScholarQuery(cluster=options.cluster_id)
    else:
        query = SearchScholarQuery()
        if options.author:
            query.set_author(options.author)
        if options.allw:
            query.set_words(options.allw)
        if options.some:
            query.set_words_some(options.some)
        if options.none:
            query.set_words_none(options.none)
        if options.phrase:
            query.set_phrase(options.phrase)
        if options.title_only:
            query.set_scope(True)
        if options.pub:
            query.set_pub(options.pub)
        if options.after or options.before:
            query.set_timeframe(options.after, options.before)
        if options.no_patents:
            query.set_include_patents(False)
        if options.no_citations:
            query.set_include_citations(False)

    if options.count is not None:
        options.count = min(options.count, ScholarConf.MAX_PAGE_RESULTS)
        query.set_num_page_results(options.count)

    querier.send_query(query)

    if options.csv:
        csv(querier)
    elif options.csv_header:
        csv(querier, header=True)
    elif options.citation is not None:
        citation_export(querier)
    else:
        txt(querier, with_globals=options.txt_globals)

    if options.cookie_file:
        querier.save_cookies()

    return 0

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import datetime, os, errno, shutil, tempfile

from .log import Log

def rm(path :str, recursive :bool = False):
    """
    Remove a file (rm -f).
    Args:
        path: A string containing the path of the file to remove.
        recursive: Pass True to remove recursively a directory (rm -rf).
    """
    if os.path.isdir(path):
        if recursive:
            shutil.rmtree(path)
        else:
            os.rmdir(path)
    else:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass

def ctime(path) -> datetime.datetime:
    """
    Retrieve the creation date of a file.
    Args:
        path: A string containing the path of the file to remove.
    Returns:
        The corresponding datetime.
    """
    posix_time = os.path.getctime(path)
    return datetime.datetime.utcfromtimestamp(posix_time)

def mtime(path) -> datetime.datetime:
    """
    Retrieve the modification date of a file.
    Args:
        path: A string containing the path of the file to remove.
    Returns:
        The corresponding datetime.
    """
    posix_time = os.path.getmtime(path)
    return datetime.datetime.utcfromtimestamp(posix_time)

def mkdir(directory):
    """
    Create a directory (mkdir -p).
    Args:
        directory: A String containing an absolute path.
    Raises:
        OSError: If the directory cannot be created.
    """
    # http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
    try:
        if not os.path.exists(directory):
            Log.info("Creating '%s' directory" % directory)
        os.makedirs(directory)
    except FileExistsError:
        pass
    except OSError as e: # Python >2.5
        if e.errno == errno.EEXIST and os.path.isdir(directory):
            pass
        else:
            raise OSError("Cannot mkdir %s: %s" % (directory, e))

def check_writable_directory(directory):
    """
    Tests whether a directory is writable.
    Args:
        directory: A String containing an absolute path.
    Raises:
        RuntimeError: If the directory does not exists or isn't writable.
    """
    if not os.path.exists(directory):
        raise RuntimeError("Directory '%s' does not exists" % directory)
    if not os.access(directory, os.W_OK | os.X_OK):
        raise RuntimeError("Directory '%s' is not writable" % directory)
    try:
        with tempfile.TemporaryFile(dir = directory):
            pass
    except Exception as e:
        raise RuntimeError("Cannot write into directory '%s': %s" % (directory, e))

def find(dir_name :str) -> list:
    """
    List regular files in a stored in given directory or one of its subdirectories.
    Args:
        dir_name: A String corresponding to an existing directory.
    Returns:
        A list of String, each of them corresponding to a file.
    """
    filenames = list()
    files = os.listdir(dir_name)
    for base_name in files:
        cur_path = os.path.join(dir_name, base_name)
        if os.path.isdir(cur_path):
            filenames += find(cur_path)
        else:
            filenames.append(cur_path)
    return filenames



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import json, os, pickle
from functools                  import partial
from minifold.cache             import DEFAULT_CACHE_STORAGE_BASE_DIR
from minifold.entries_connector import EntriesConnector
from minifold.filesystem        import mkdir
from minifold.cache             import JsonCacheConnector
from minifold.log               import Log

class CachedEntriesConnector(EntriesConnector):
    def __init__(
        self,
        load_entries,
        cache_filename :str,
        load_cache,
        save_cache,
        read_mode,
        write_mode,
        with_cache :bool = True
    ):
        loaded_from_cache = False
        if with_cache:
            try:
                with open(cache_filename, read_mode) as f:
                    Log.info("%s: Loading cache from [%s]" % (type(self), cache_filename))
                    entries = load_cache(f)
                    Log.info("Loaded %d entries" % len(entries))
                    loaded_from_cache = True
            except FileNotFoundError:
                Log.debug("%s: Cache [%s] not found" % (type(self), cache_filename))
                pass
            except Exception as e:
                Log.debug("%s: Cache [%s] corrupted" % (type(self), cache_filename))
                Log.error(e)
                pass

        # Parse the input data (if needed)
        if not loaded_from_cache:
            entries = load_entries()
            Log.info("Loaded %d entries" % len(entries))

        # Save into cache (if needed)
        if with_cache and not loaded_from_cache:
            Log.info("%s: Saving data into cache [%s]" % (type(self), cache_filename))
            mkdir(os.path.dirname(cache_filename))
            with open(cache_filename, write_mode) as f:
                save_cache(entries, f)

        super().__init__(entries)

class JsonCachedConnector(CachedEntriesConnector):
    def __init__(self, load_entries, cache_filename :str, **kwargs):
        super().__init__(
            load_entries,
            cache_filename,
            load_cache = json.load,
            save_cache = partial(json.dump, indent=4),
            read_mode  = "r",
            write_mode = "w",
            **kwargs
        )

class PickleCachedConnector(CachedEntriesConnector):
    def __init__(self, load_entries, cache_filename :str, **kwargs):
        super().__init__(
            load_entries,
            cache_filename,
            load_cache = pickle.load,
            save_cache = pickle.dump,
            read_mode  = "rb",
            write_mode = "wb",
            **kwargs
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from enum import IntEnum

class DocType(IntEnum):
    BOOKS_AND_THESES = 0 # Books and Theses
    ARTICLE          = 1 # ART, Conference and Workshop Papers
    COMMUNICATION    = 2 # COMM
    JOURNAL          = 3 # Journal Articles
    REPORT           = 4 # Report, Informal Publications
    POSTER           = 5 # POSTER
    HDR              = 6 # HDR
    CHAPTER          = 7 # COUV, Editorship
    PATENT           = 8 # PATENT
    UNKNOWN          = 9999 # UNDEFINED or ""

    def __str__(self) -> str:
        s = IntEnum.__str__(self)
        class_name = "%s" % self.__class__.__name__
        s = s[len(class_name)+1:]
        s = s.replace("_", " ")
        return s.title()

    def __lt__(self, other) -> bool:
        return self.value < other.value

def doc_type_to_html(t :DocType, n :int = 1) -> str:
    # Default %s is not wonderful
    if   t == DocType.UNKNOWN: ret = "Misc"
    elif t == DocType.CHAPTER: ret = "Editorship"
    elif t == DocType.JOURNAL: ret = "Journal article"
    else: ret = "%s" % t

    # Add "s" at the end if needed
    if n > 1 and t not in [DocType.BOOKS_AND_THESES, DocType.HDR, DocType.CHAPTER, DocType.UNKNOWN]:
        ret = "%ss" % ret

    return ret





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .query                 import Query, ACTION_READ

def count_gen(gen) -> int:
    """
    Count the number of elements listed by a generator.
    Args:
        gen: A generator.
    Returns:
        The number of elements listed by a generator.
    """
    n = 0
    for x in gen: n += 1
    return n

def count(entries :list) -> int:
    """
    Count the number of entries.
    Args:
        entries: A list of entries
    Returns:
        The number of element in entries
    """
    return len(entries)

class CountConnector(Connector):
    def __init__(self, child :Connector):
        super().__init__()
        self.m_child = child

    def query(self, query :Query) -> int:
        super().query(query)
        if query.action != ACTION_READ:
            raise ValueError("Invalid action in query %s" % query)
        return count(self.m_child.query(query))
